# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for repository commit builder."""

import os

from breezy import errors, osutils, repository, tests
from breezy import revision as _mod_revision
from breezy.bzr import inventorytree
from breezy.bzr.inventorytree import InventoryTreeChange
from breezy.tests import features, per_repository

from ..test_bedding import override_whoami


class TestCommitBuilder(per_repository.TestCaseWithRepository):
    def test_get_commit_builder(self):
        branch = self.make_branch(".")
        branch.repository.lock_write()
        builder = branch.repository.get_commit_builder(
            branch, [], branch.get_config_stack()
        )
        self.assertIsInstance(builder, repository.CommitBuilder)
        self.assertTrue(builder.random_revid)
        branch.repository.commit_write_group()
        branch.repository.unlock()

    def test_finish_inventory_record_iter_changes(self):
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            builder = tree.branch.get_commit_builder([])
            try:
                list(
                    builder.record_iter_changes(
                        tree, tree.last_revision(), tree.iter_changes(tree.basis_tree())
                    )
                )
                builder.finish_inventory()
            except:
                builder.abort()
                raise
            repo = tree.branch.repository
            repo.commit_write_group()

    def test_abort_record_iter_changes(self):
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            builder = tree.branch.get_commit_builder([])
            try:
                basis = tree.basis_tree()
                last_rev = tree.last_revision()
                changes = tree.iter_changes(basis)
                list(builder.record_iter_changes(tree, last_rev, changes))
                builder.finish_inventory()
            finally:
                builder.abort()

    def test_commit_lossy(self):
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            builder = tree.branch.get_commit_builder([], lossy=True)
            list(
                builder.record_iter_changes(
                    tree, tree.last_revision(), tree.iter_changes(tree.basis_tree())
                )
            )
            builder.finish_inventory()
            rev_id = builder.commit("foo bar blah")
        rev = tree.branch.repository.get_revision(rev_id)
        self.assertEqual("foo bar blah", rev.message)

    def test_commit_message(self):
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            builder = tree.branch.get_commit_builder([])
            list(
                builder.record_iter_changes(
                    tree, tree.last_revision(), tree.iter_changes(tree.basis_tree())
                )
            )
            builder.finish_inventory()
            rev_id = builder.commit("foo bar blah")
        rev = tree.branch.repository.get_revision(rev_id)
        self.assertEqual("foo bar blah", rev.message)

    def test_updates_branch(self):
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            builder = tree.branch.get_commit_builder([])
            list(
                builder.record_iter_changes(
                    tree, tree.last_revision(), tree.iter_changes(tree.basis_tree())
                )
            )
            builder.finish_inventory()
            will_update_branch = builder.updates_branch
            rev_id = builder.commit("might update the branch")
        actually_updated_branch = tree.branch.last_revision() == rev_id
        self.assertEqual(actually_updated_branch, will_update_branch)

    def test_commit_with_revision_id_record_iter_changes(self):
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            # use a unicode revision id to test more corner cases.
            # The repository layer is meant to handle this.
            revision_id = "\xc8abc".encode()
            try:
                try:
                    builder = tree.branch.get_commit_builder(
                        [], revision_id=revision_id
                    )
                except errors.NonAsciiRevisionId:
                    revision_id = b"abc"
                    builder = tree.branch.get_commit_builder(
                        [], revision_id=revision_id
                    )
            except repository.CannotSetRevisionId:
                # This format doesn't support supplied revision ids
                return
            self.assertFalse(builder.random_revid)
            try:
                list(
                    builder.record_iter_changes(
                        tree, tree.last_revision(), tree.iter_changes(tree.basis_tree())
                    )
                )
                builder.finish_inventory()
            except:
                builder.abort()
                raise
            self.assertEqual(revision_id, builder.commit("foo bar"))
        self.assertTrue(tree.branch.repository.has_revision(revision_id))
        # the revision id must be set on the inventory when saving it. This
        # does not precisely test that - a repository that wants to can add it
        # on deserialisation, but thats all the current contract guarantees
        # anyway.
        self.assertEqual(
            revision_id,
            tree.branch.repository.revision_tree(revision_id).get_revision_id(),
        )

    def test_commit_without_root_errors(self):
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            builder = tree.branch.get_commit_builder([])

            def do_commit():
                try:
                    list(builder.record_iter_changes(tree, tree.last_revision(), []))
                    builder.finish_inventory()
                except:
                    builder.abort()
                    raise
                else:
                    builder.commit("msg")

            self.assertRaises(errors.RootMissing, do_commit)

    def test_commit_unchanged_root_record_iter_changes(self):
        tree = self.make_branch_and_tree(".")
        old_revision_id = tree.commit("oldrev")
        tree.lock_write()
        builder = tree.branch.get_commit_builder([old_revision_id])
        try:
            list(builder.record_iter_changes(tree, old_revision_id, []))
            # Regardless of repository root behaviour we should consider this a
            # pointless commit.
            self.assertFalse(builder.any_changes())
            builder.finish_inventory()
            builder.commit("rev")
            builder_tree = builder.revision_tree()
            new_root_revision = builder_tree.get_file_revision("")
            if tree.branch.repository.supports_rich_root():
                # We should not have seen a new root revision
                self.assertEqual(old_revision_id, new_root_revision)
            else:
                # We should see a new root revision
                self.assertNotEqual(old_revision_id, new_root_revision)
        finally:
            tree.unlock()

    def test_record_delete_record_iter_changes(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["foo"])
        tree.add(["foo"])
        foo_id = tree.path2id("foo")
        rev_id = tree.commit("added foo")
        with tree.lock_write():
            builder = tree.branch.get_commit_builder([rev_id])
            try:
                delete_change = InventoryTreeChange(
                    foo_id,
                    ("foo", None),
                    True,
                    (True, False),
                    (tree.path2id(""), None),
                    ("foo", None),
                    ("file", None),
                    (False, None),
                )
                list(builder.record_iter_changes(tree, rev_id, [delete_change]))
                self.assertEqual(
                    ("foo", None, foo_id, None), builder.get_basis_delta()[0]
                )
                self.assertTrue(builder.any_changes())
                builder.finish_inventory()
                builder.commit("delete foo")
            except:
                builder.abort()
                raise
        rev_tree = builder.revision_tree()
        rev_tree.lock_read()
        self.addCleanup(rev_tree.unlock)
        self.assertFalse(rev_tree.is_versioned("foo"))

    def test_revision_tree_record_iter_changes(self):
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            builder = tree.branch.get_commit_builder([])
            try:
                list(
                    builder.record_iter_changes(
                        tree,
                        _mod_revision.NULL_REVISION,
                        tree.iter_changes(tree.basis_tree()),
                    )
                )
                builder.finish_inventory()
                rev_id = builder.commit("foo bar")
            except:
                builder.abort()
                raise
            rev_tree = builder.revision_tree()
            # Just a couple simple tests to ensure that it actually follows
            # the RevisionTree api.
            self.assertEqual(rev_id, rev_tree.get_revision_id())
            self.assertEqual((), tuple(rev_tree.get_parent_ids()))

    def test_root_entry_has_revision(self):
        # test the root revision created and put in the basis
        # has the right rev id.
        # XXX: RBC 20081118 - this test is too big, it depends on the exact
        # behaviour of tree methods and so on; it should be written to the
        # commit builder interface directly.
        tree = self.make_branch_and_tree(".")
        rev_id = tree.commit("message")
        basis_tree = tree.basis_tree()
        basis_tree.lock_read()
        self.addCleanup(basis_tree.unlock)
        self.assertEqual(rev_id, basis_tree.get_file_revision(""))

    def _get_revtrees(self, tree, revision_ids):
        with tree.lock_read():
            trees = list(tree.branch.repository.revision_trees(revision_ids))
            for _tree in trees:
                _tree.lock_read()
                self.addCleanup(_tree.unlock)
            return trees

    def test_last_modified_revision_after_commit_root_unchanged(self):
        # commiting without changing the root does not change the
        # last modified except on non-rich-root-repositories.
        tree = self.make_branch_and_tree(".")
        rev1 = tree.commit("rev1")
        rev2 = tree.commit("rev2")
        tree1, tree2 = self._get_revtrees(tree, [rev1, rev2])
        self.assertEqual(rev1, tree1.get_file_revision(""))
        if tree.branch.repository.supports_rich_root():
            self.assertEqual(rev1, tree2.get_file_revision(""))
        else:
            self.assertEqual(rev2, tree2.get_file_revision(""))

    def _add_commit_check_unchanged(self, tree, name):
        tree.add([name])
        if tree.supports_file_ids:
            file_id = tree.path2id(name)
        else:
            file_id = None
        self._commit_check_unchanged(tree, name, file_id)

    def _commit_check_unchanged(self, tree, name, file_id):
        rev1 = tree.commit("rev1")
        rev2 = self.mini_commit_record_iter_changes(tree, name, name, False, False)
        tree1, tree2 = self._get_revtrees(tree, [rev1, rev2])
        self.assertEqual(rev1, tree1.get_file_revision(name))
        self.assertEqual(rev1, tree2.get_file_revision(name))
        if tree.supports_file_ids:
            expected_graph = {}
            expected_graph[(file_id, rev1)] = ()
            self.assertFileGraph(expected_graph, tree, (file_id, rev1))

    def test_last_modified_revision_after_commit_dir_unchanged(self):
        # committing without changing a dir does not change the last modified.
        tree = self.make_branch_and_tree(".")
        if not tree.has_versioned_directories():
            raise tests.TestNotApplicable(
                "Format does not support versioned directories"
            )
        self.build_tree(["dir/"])
        self._add_commit_check_unchanged(tree, "dir")

    def test_last_modified_revision_after_commit_dir_contents_unchanged(self):
        # committing without changing a dir does not change the last modified
        # of the dir even the dirs contents are changed.
        tree = self.make_branch_and_tree(".")
        self.build_tree(["dir/", "dir/orig"])
        tree.add(["dir", "dir/orig"])
        rev1 = tree.commit("rev1")
        self.build_tree(["dir/content"])
        tree.add(["dir/content"])
        rev2 = tree.commit("rev2")
        tree1, tree2 = self._get_revtrees(tree, [rev1, rev2])
        self.assertEqual(rev1, tree1.get_file_revision("dir"))
        self.assertEqual(rev1, tree2.get_file_revision("dir"))
        if tree.supports_file_ids:
            dir_id = tree1.path2id("dir")
            expected_graph = {(dir_id, rev1): ()}
            self.assertFileGraph(expected_graph, tree, (dir_id, rev1))

    def test_last_modified_revision_after_commit_file_unchanged(self):
        # committing without changing a file does not change the last modified.
        tree = self.make_branch_and_tree(".")
        self.build_tree(["file"])
        self._add_commit_check_unchanged(tree, "file")

    def test_last_modified_revision_after_commit_link_unchanged(self):
        # committing without changing a link does not change the last modified.
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.make_branch_and_tree(".")
        os.symlink("target", "link")
        self._add_commit_check_unchanged(tree, "link")

    def test_last_modified_revision_after_commit_reference_unchanged(self):
        # committing without changing a subtree does not change the last
        # modified.
        tree = self.make_branch_and_tree(".")
        subtree = self.make_reference("reference")
        subtree.commit("")
        try:
            tree.add_reference(subtree)
            self._commit_check_unchanged(
                tree,
                "reference",
                subtree.path2id("") if subtree.supports_file_ids else None,
            )
        except errors.UnsupportedOperation:
            return

    def _add_commit_renamed_check_changed(self, tree, name, expect_fs_hash=False):
        def rename():
            tree.rename_one(name, "new_" + name)

        self._add_commit_change_check_changed(
            tree, (name, "new_" + name), rename, expect_fs_hash=expect_fs_hash
        )

    def _commit_renamed_check_changed(self, tree, name, expect_fs_hash=False):
        def rename():
            tree.rename_one(name, "new_" + name)

        self._commit_change_check_changed(
            tree, [name, "new_" + name], rename, expect_fs_hash=expect_fs_hash
        )

    def test_last_modified_revision_after_rename_dir_changes(self):
        # renaming a dir changes the last modified.
        tree = self.make_branch_and_tree(".")
        if not tree.has_versioned_directories():
            raise tests.TestNotApplicable(
                "Format does not support versioned directories"
            )
        self.build_tree(["dir/"])
        self._add_commit_renamed_check_changed(tree, "dir")

    def test_last_modified_revision_after_rename_file_changes(self):
        # renaming a file changes the last modified.
        tree = self.make_branch_and_tree(".")
        self.build_tree(["file"])
        self._add_commit_renamed_check_changed(tree, "file", expect_fs_hash=True)

    def test_last_modified_revision_after_rename_link_changes(self):
        # renaming a link changes the last modified.
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.make_branch_and_tree(".")
        os.symlink("target", "link")
        self._add_commit_renamed_check_changed(tree, "link")

    def test_last_modified_revision_after_rename_ref_changes(self):
        # renaming a reference changes the last modified.
        tree = self.make_branch_and_tree(".")
        subtree = self.make_reference("reference")
        subtree.commit("")
        try:
            tree.add_reference(subtree)
            self._commit_renamed_check_changed(tree, "reference")
        except errors.UnsupportedOperation:
            return

    def _add_commit_reparent_check_changed(self, tree, name, expect_fs_hash=False):
        self.build_tree(["newparent/"])
        tree.add(["newparent"])

        def reparent():
            tree.rename_one(name, "newparent/new_" + name)

        self._add_commit_change_check_changed(
            tree,
            (name, "newparent/new_" + name),
            reparent,
            expect_fs_hash=expect_fs_hash,
        )

    def test_last_modified_revision_after_reparent_dir_changes(self):
        # reparenting a dir changes the last modified.
        tree = self.make_branch_and_tree(".")
        if not tree.has_versioned_directories():
            raise tests.TestNotApplicable(
                "Format does not support versioned directories"
            )
        self.build_tree(["dir/"])
        self._add_commit_reparent_check_changed(tree, "dir")

    def test_last_modified_revision_after_reparent_file_changes(self):
        # reparenting a file changes the last modified.
        tree = self.make_branch_and_tree(".")
        self.build_tree(["file"])
        self._add_commit_reparent_check_changed(tree, "file", expect_fs_hash=True)

    def test_last_modified_revision_after_reparent_link_changes(self):
        # reparenting a link changes the last modified.
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.make_branch_and_tree(".")
        os.symlink("target", "link")
        self._add_commit_reparent_check_changed(tree, "link")

    def _add_commit_change_check_changed(
        self, tree, names, changer, expect_fs_hash=False
    ):
        tree.add([names[0]])
        self.assertTrue(tree.is_versioned(names[0]))
        self._commit_change_check_changed(
            tree, names, changer, expect_fs_hash=expect_fs_hash
        )

    def _commit_change_check_changed(self, tree, names, changer, expect_fs_hash=False):
        rev1 = tree.commit("rev1")
        changer()
        rev2 = self.mini_commit_record_iter_changes(
            tree, names[0], names[1], expect_fs_hash=expect_fs_hash
        )
        tree1, tree2 = self._get_revtrees(tree, [rev1, rev2])
        self.assertEqual(rev1, tree1.get_file_revision(names[0]))
        self.assertEqual(rev2, tree2.get_file_revision(names[1]))
        if tree1.supports_file_ids:
            file_id = tree1.path2id(names[0])
            expected_graph = {}
            expected_graph[(file_id, rev1)] = ()
            expected_graph[(file_id, rev2)] = ((file_id, rev1),)
            self.assertFileGraph(expected_graph, tree, (file_id, rev2))

    def mini_commit_record_iter_changes(
        self,
        tree,
        name,
        new_name,
        records_version=True,
        delta_against_basis=True,
        expect_fs_hash=False,
    ):
        """Perform a miniature commit looking for record entry results.

        This version uses the record_iter_changes interface.

        :param tree: The tree to commit.
        :param name: The path in the basis tree of the tree being committed.
        :param new_name: The path in the tree being committed.
        :param records_version: True if the commit of new_name is expected to
            record a new version.
        :param delta_against_basis: True of the commit of new_name is expected
            to have a delta against the basis.
        :param expect_fs_hash: If true, looks for a fs hash output from
            record_iter_changes.
        """
        with tree.lock_write():
            # mini manual commit here so we can check the return of
            # record_iter_changes
            parent_ids = tree.get_parent_ids()
            builder = tree.branch.get_commit_builder(parent_ids)
            try:
                parent_tree = tree.basis_tree()
                with parent_tree.lock_read():
                    changes = list(tree.iter_changes(parent_tree))
                result = list(builder.record_iter_changes(tree, parent_ids[0], changes))
                self.assertTrue(tree.is_versioned(new_name))
                if isinstance(tree, inventorytree.InventoryTree):
                    file_id = tree.path2id(new_name)
                    self.assertIsNot(None, file_id)
                    # TODO(jelmer): record_iter_changes shouldn't yield
                    # data that is WorkingTree-format-specific and uses file ids.
                    if expect_fs_hash:
                        tree_file_stat = tree.get_file_with_stat(new_name)
                        tree_file_stat[0].close()
                        self.assertLength(1, result)
                        result = result[0]
                        self.assertEqual(result[0], new_name)
                        self.assertEqual(result[1][0], tree.get_file_sha1(new_name))
                        self.assertEqualStat(result[1][1], tree_file_stat[1])
                    else:
                        self.assertEqual([], result)
                builder.finish_inventory()
                if tree.branch.repository._format.supports_full_versioned_files:
                    inv_key = (builder._new_revision_id,)
                    inv_sha1 = tree.branch.repository.inventories.get_sha1s([inv_key])[
                        inv_key
                    ]
                    self.assertEqual(inv_sha1, builder.inv_sha1)
                rev2 = builder.commit("rev2")
            except BaseException:
                builder.abort()
                raise
            delta = builder.get_basis_delta()
            delta_dict = {change[1]: change for change in delta}
            if tree.branch.repository._format.records_per_file_revision:
                version_recorded = (
                    new_name in delta_dict
                    and delta_dict[new_name][3] is not None
                    and delta_dict[new_name][3].revision == rev2
                )
                if records_version:
                    self.assertTrue(version_recorded)
                else:
                    self.assertFalse(version_recorded)

            revtree = builder.revision_tree()
            new_entry = next(revtree.iter_entries_by_dir(specific_files=[new_name]))[1]

            if delta_against_basis:
                (delta_old_name, delta_new_name, delta_file_id, delta_entry) = (
                    delta_dict[new_name]
                )
                self.assertEqual(delta_new_name, new_name)
                if tree.supports_rename_tracking():
                    self.assertEqual(name, delta_old_name)
                else:
                    self.assertIn(delta_old_name, (name, None))
                if tree.supports_setting_file_ids():
                    self.assertEqual(delta_file_id, file_id)
                    self.assertEqual(delta_entry.file_id, file_id)
                self.assertEqual(delta_entry.kind, new_entry.kind)
                self.assertEqual(delta_entry.name, new_entry.name)
                self.assertEqual(delta_entry.parent_id, new_entry.parent_id)
                if delta_entry.kind == "file":
                    self.assertEqual(
                        delta_entry.text_size, revtree.get_file_size(new_name)
                    )
                    if getattr(delta_entry, "text_sha1", None):
                        self.assertEqual(
                            delta_entry.text_sha1, revtree.get_file_sha1(new_name)
                        )
                elif delta_entry.kind == "symlink":
                    self.assertEqual(
                        delta_entry.symlink_target, new_entry.symlink_target
                    )
            else:
                if tree.branch.repository._format.records_per_file_revision:
                    self.assertFalse(version_recorded)
            tree.set_parent_ids([rev2])
        return rev2

    def assertFileGraph(self, expected_graph, tree, tip):
        # all the changes that have occured should be in the ancestry
        # (closest to a public per-file graph API we have today)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        g = dict(tree.branch.repository.get_file_graph().iter_ancestry([tip]))
        self.assertEqual(expected_graph, g)

    def test_last_modified_revision_after_content_file_changes(self):
        # altering a file changes the last modified.
        tree = self.make_branch_and_tree(".")
        self.build_tree(["file"])

        def change_file():
            tree.put_file_bytes_non_atomic("file", b"new content")

        self._add_commit_change_check_changed(
            tree, ("file", "file"), change_file, expect_fs_hash=True
        )

    def _test_last_mod_rev_after_content_link_changes(self, link, target, newtarget):
        # changing a link changes the last modified.
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.make_branch_and_tree(".")
        os.symlink(target, link)

        def change_link():
            os.unlink(link)
            os.symlink(newtarget, link)

        self._add_commit_change_check_changed(tree, (link, link), change_link)

    def test_last_modified_rev_after_content_link_changes(self):
        self._test_last_mod_rev_after_content_link_changes(
            "link", "target", "newtarget"
        )

    def test_last_modified_rev_after_content_unicode_link_changes(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        self._test_last_mod_rev_after_content_link_changes(
            "li\u1234nk", "targ\N{EURO SIGN}t", "n\N{EURO SIGN}wtarget"
        )

    def _commit_sprout(self, tree, name):
        tree.add([name])
        rev_id = tree.commit("rev")
        return rev_id, tree.controldir.sprout("t2").open_workingtree()

    def _rename_in_tree(self, tree, name, message):
        tree.rename_one(name, "new_" + name)
        return tree.commit(message)

    def _commit_sprout_rename_merge(self, tree1, name, expect_fs_hash=False):
        """Do a rename in both trees."""
        rev1, tree2 = self._commit_sprout(tree1, name)
        if tree2.supports_file_ids:
            file_id = tree2.path2id(name)
            self.assertIsNot(None, file_id)
        self.assertTrue(tree2.is_versioned(name))
        # change both sides equally
        rev2 = self._rename_in_tree(tree1, name, "rev2")
        rev3 = self._rename_in_tree(tree2, name, "rev3")
        tree1.merge_from_branch(tree2.branch)
        rev4 = self.mini_commit_record_iter_changes(
            tree1,
            "new_" + name,
            "new_" + name,
            expect_fs_hash=expect_fs_hash,
            delta_against_basis=tree1.supports_rename_tracking(),
        )
        (tree3,) = self._get_revtrees(tree1, [rev4])
        if tree1.supports_file_ids:
            expected_graph = {}
            if tree1.supports_rename_tracking():
                self.assertEqual(rev4, tree3.get_file_revision("new_" + name))
                expected_graph[(file_id, rev1)] = ()
                expected_graph[(file_id, rev2)] = ((file_id, rev1),)
                expected_graph[(file_id, rev3)] = ((file_id, rev1),)
                expected_graph[(file_id, rev4)] = (
                    (file_id, rev2),
                    (file_id, rev3),
                )
            else:
                self.assertEqual(rev2, tree3.get_file_revision("new_" + name))
                expected_graph[(file_id, rev4)] = ()
            self.assertFileGraph(expected_graph, tree1, (file_id, rev4))

    def test_last_modified_revision_after_merge_dir_changes(self):
        # merge a dir changes the last modified.
        tree1 = self.make_branch_and_tree("t1")
        if not tree1.has_versioned_directories():
            raise tests.TestNotApplicable(
                "Format does not support versioned directories"
            )
        self.build_tree(["t1/dir/"])
        self._commit_sprout_rename_merge(tree1, "dir")

    def test_last_modified_revision_after_merge_file_changes(self):
        # merge a file changes the last modified.
        tree1 = self.make_branch_and_tree("t1")
        self.build_tree(["t1/file"])
        self._commit_sprout_rename_merge(tree1, "file", expect_fs_hash=True)

    def test_last_modified_revision_after_merge_link_changes(self):
        # merge a link changes the last modified.
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree1 = self.make_branch_and_tree("t1")
        os.symlink("target", "t1/link")
        self._commit_sprout_rename_merge(tree1, "link")

    def _commit_sprout_rename_merge_converged(self, tree1, name):
        # Make a merge which just incorporates a change from a branch:
        # The per-file graph is straight line, and no alteration occurs
        # in the inventory.
        # Part 1: change in the merged branch.
        rev1, tree2 = self._commit_sprout(tree1, name)
        if tree2.supports_file_ids:
            file_id = tree2.path2id(name)
            self.assertIsNot(None, file_id)
        # change on the other side to merge back
        rev2 = self._rename_in_tree(tree2, name, "rev2")
        tree1.merge_from_branch(tree2.branch)

        if tree2.supports_file_ids:

            def _check_graph(in_tree, changed_in_tree):
                self.mini_commit_record_iter_changes(
                    in_tree,
                    name,
                    "new_" + name,
                    False,
                    delta_against_basis=changed_in_tree,
                )
                (tree3,) = self._get_revtrees(in_tree, [rev2])
                self.assertEqual(rev2, tree3.get_file_revision("new_" + name))
                expected_graph = {}
                expected_graph[(file_id, rev1)] = ()
                expected_graph[(file_id, rev2)] = ((file_id, rev1),)
                self.assertFileGraph(expected_graph, in_tree, (file_id, rev2))

            _check_graph(tree1, True)
        # Part 2: change in the merged into branch - we use tree2 that has a
        # change to name, branch tree1 and give it an unrelated change, then
        # merge that to t2.
        other_tree = tree1.controldir.sprout("t3").open_workingtree()
        other_tree.commit("other_rev")
        tree2.merge_from_branch(other_tree.branch)
        if tree2.supports_file_ids:
            _check_graph(tree2, False)

    def _commit_sprout_make_merge(self, tree1, make):
        # Make a merge which incorporates the addition of a new object to
        # another branch. The per-file graph shows no additional change
        # in the merge because its a straight line.
        tree1.commit("rev1")
        tree2 = tree1.controldir.sprout("t2").open_workingtree()
        # make and commit on the other side to merge back
        make("t2/name")
        tree2.add(["name"])
        self.assertTrue(tree2.is_versioned("name"))
        rev2 = tree2.commit("rev2")
        tree1.merge_from_branch(tree2.branch)
        self.mini_commit_record_iter_changes(tree1, None, "name", False)
        (tree3,) = self._get_revtrees(tree1, [rev2])
        # in rev2, name should be only changed in rev2
        self.assertEqual(rev2, tree3.get_file_revision("name"))
        if tree2.supports_file_ids:
            file_id = tree2.path2id("name")
            expected_graph = {}
            expected_graph[(file_id, rev2)] = ()
            self.assertFileGraph(expected_graph, tree1, (file_id, rev2))

    def test_last_modified_revision_after_converged_merge_dir_unchanged(self):
        # merge a dir that changed preserves the last modified.
        tree1 = self.make_branch_and_tree("t1")
        if not tree1.has_versioned_directories():
            raise tests.TestNotApplicable(
                "Format does not support versioned directories"
            )
        self.build_tree(["t1/dir/"])
        self._commit_sprout_rename_merge_converged(tree1, "dir")

    def test_last_modified_revision_after_converged_merge_file_unchanged(self):
        # merge a file that changed preserves the last modified.
        tree1 = self.make_branch_and_tree("t1")
        self.build_tree(["t1/file"])
        self._commit_sprout_rename_merge_converged(tree1, "file")

    def test_last_modified_revision_after_converged_merge_link_unchanged(self):
        # merge a link that changed preserves the last modified.
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree1 = self.make_branch_and_tree("t1")
        os.symlink("target", "t1/link")
        self._commit_sprout_rename_merge_converged(tree1, "link")

    def test_last_modified_revision_after_merge_new_dir_unchanged(self):
        # merge a new dir does not change the last modified.
        tree1 = self.make_branch_and_tree("t1")
        if not tree1.has_versioned_directories():
            raise tests.TestNotApplicable(
                "Format does not support versioned directories"
            )
        self._commit_sprout_make_merge(tree1, self.make_dir)

    def test_last_modified_revision_after_merge_new_file_unchanged(self):
        # merge a new file does not change the last modified.
        tree1 = self.make_branch_and_tree("t1")
        self._commit_sprout_make_merge(tree1, self.make_file)

    def test_last_modified_revision_after_merge_new_link_unchanged(self):
        # merge a new link does not change the last modified.
        tree1 = self.make_branch_and_tree("t1")
        self._commit_sprout_make_merge(tree1, self.make_link)

    def make_dir(self, name):
        self.build_tree([name + "/"])

    def make_file(self, name):
        self.build_tree([name])

    def make_link(self, name):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        os.symlink("target", name)

    def make_reference(self, name):
        tree = self.make_branch_and_tree(name)
        if not tree.branch.repository._format.rich_root_data:
            raise tests.TestNotApplicable("format does not support rich roots")
        tree.commit("foo")
        return tree

    def _check_kind_change(self, make_before, make_after, expect_fs_hash=False):
        tree = self.make_branch_and_tree(".")
        path = "name"
        make_before(path)

        def change_kind():
            if osutils.file_kind(path) == "directory":
                osutils.rmtree(path)
            else:
                osutils.delete_any(path)
            make_after(path)

        self._add_commit_change_check_changed(
            tree, (path, path), change_kind, expect_fs_hash=expect_fs_hash
        )

    def test_last_modified_dir_file(self):
        if not self.repository_format.supports_versioned_directories:
            # TODO(jelmer): Perhaps test this by creating a directory
            # with a file in it?
            raise tests.TestNotApplicable(
                "format does not support versioned directories"
            )
        try:
            self._check_kind_change(self.make_dir, self.make_file, expect_fs_hash=True)
        except errors.UnsupportedKindChange:
            raise tests.TestSkipped(
                "tree does not support changing entry kind from directory to file"
            )

    def test_last_modified_dir_link(self):
        if not self.repository_format.supports_versioned_directories:
            # TODO(jelmer): Perhaps test this by creating a directory
            # with a file in it?
            raise tests.TestNotApplicable(
                "format does not support versioned directories"
            )
        try:
            self._check_kind_change(self.make_dir, self.make_link)
        except errors.UnsupportedKindChange:
            raise tests.TestSkipped(
                "tree does not support changing entry kind from directory to link"
            )

    def test_last_modified_link_file(self):
        self._check_kind_change(self.make_link, self.make_file, expect_fs_hash=True)

    def test_last_modified_link_dir(self):
        if not self.repository_format.supports_versioned_directories:
            # TODO(jelmer): Perhaps test this by creating a directory
            # with a file in it?
            raise tests.TestNotApplicable(
                "format does not support versioned directories"
            )

        self._check_kind_change(self.make_link, self.make_dir)

    def test_last_modified_file_dir(self):
        if not self.repository_format.supports_versioned_directories:
            # TODO(jelmer): Perhaps test this by creating a directory
            # with a file in it?
            raise tests.TestNotApplicable(
                "format does not support versioned directories"
            )

        self._check_kind_change(self.make_file, self.make_dir)

    def test_last_modified_file_link(self):
        self._check_kind_change(self.make_file, self.make_link)

    def test_get_commit_builder_with_invalid_revprops(self):
        branch = self.make_branch(".")
        branch.repository.lock_write()
        self.addCleanup(branch.repository.unlock)
        self.assertRaises(
            ValueError,
            branch.repository.get_commit_builder,
            branch,
            [],
            branch.get_config_stack(),
            revprops={"invalid": "property\rwith\r\ninvalid chars"},
        )

    def test_get_commit_builder_with_surrogateescape(self):
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            builder = tree.branch.get_commit_builder(
                [],
                revprops={
                    "invalid": "property" + b"\xc0".decode("utf-8", "surrogateescape")
                },
            )
            list(
                builder.record_iter_changes(
                    tree, tree.last_revision(), tree.iter_changes(tree.basis_tree())
                )
            )
            builder.finish_inventory()
            try:
                rev_id = builder.commit("foo bar blah")
            except NotImplementedError:
                raise tests.TestNotApplicable(
                    "Format does not support revision properties"
                )
        rev = tree.branch.repository.get_revision(rev_id)
        self.assertEqual("foo bar blah", rev.message)

    def test_commit_builder_commit_with_invalid_message(self):
        branch = self.make_branch(".")
        branch.repository.lock_write()
        self.addCleanup(branch.repository.unlock)
        builder = branch.repository.get_commit_builder(
            branch, [], branch.get_config_stack()
        )
        self.addCleanup(branch.repository.abort_write_group)
        self.assertRaises(ValueError, builder.commit, "Invalid\r\ncommit message\r\n")

    def test_non_ascii_str_committer_rejected(self):
        """Ensure an error is raised on a non-ascii byte string committer."""
        branch = self.make_branch(".")
        branch.repository.lock_write()
        self.addCleanup(branch.repository.unlock)
        self.assertRaises(
            UnicodeDecodeError,
            branch.repository.get_commit_builder,
            branch,
            [],
            branch.get_config_stack(),
            committer=b"Erik B\xe5gfors <erik@example.com>",
        )

    def test_stacked_repositories_reject_commit_builder(self):
        # As per bug 375013, committing to stacked repositories is currently
        # broken if we aren't in a chk repository. So old repositories with
        # fallbacks refuse to hand out a commit builder.
        repo_basis = self.make_repository("basis")
        branch = self.make_branch("local")
        repo_local = branch.repository
        try:
            repo_local.add_fallback_repository(repo_basis)
        except errors.UnstackableRepositoryFormat:
            raise tests.TestNotApplicable("not a stackable format.")
        self.addCleanup(repo_local.lock_write().unlock)
        if not repo_local._format.supports_chks:
            self.assertRaises(
                errors.BzrError,
                repo_local.get_commit_builder,
                branch,
                [],
                branch.get_config_stack(),
            )
        else:
            builder = repo_local.get_commit_builder(
                branch, [], branch.get_config_stack()
            )
            builder.abort()

    def test_committer_no_username(self):
        # Ensure that when no username is available but a committer is
        # supplied, commit works.
        override_whoami(self)
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            # Make sure no username is available.
            self.assertRaises(errors.NoWhoami, tree.branch.get_commit_builder, [])
            builder = tree.branch.get_commit_builder([], committer="me@example.com")
            try:
                list(
                    builder.record_iter_changes(
                        tree, tree.last_revision(), tree.iter_changes(tree.basis_tree())
                    )
                )
                builder.finish_inventory()
            except:
                builder.abort()
                raise
            repo = tree.branch.repository
            repo.commit_write_group()



# Copyright (C) 2009, 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy.export."""

import gzip
import os
import tarfile
import time
import zipfile
from io import BytesIO

from .. import errors, export, tests
from ..archive.tar import tarball_generator
from ..export import get_root_name
from . import features


class TestDirExport(tests.TestCaseWithTransport):
    def test_missing_file(self):
        self.build_tree(["a/", "a/b", "a/c"])
        wt = self.make_branch_and_tree(".")
        wt.add(["a", "a/b", "a/c"])
        os.unlink("a/c")
        export.export(wt, "target", format="dir")
        self.assertPathExists("target/a/b")
        self.assertPathDoesNotExist("target/a/c")

    def test_empty(self):
        wt = self.make_branch_and_tree(".")
        export.export(wt, "target", format="dir")
        self.assertEqual([], os.listdir("target"))

    def test_symlink(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        wt = self.make_branch_and_tree(".")
        os.symlink("source", "link")
        wt.add(["link"])
        export.export(wt, "target", format="dir")
        self.assertPathExists("target/link")

    def test_nested_tree(self):
        wt = self.make_branch_and_tree(".", format="development-subtree")
        subtree = self.make_branch_and_tree("subtree")
        self.build_tree(["subtree/file"])
        subtree.add(["file"])
        wt.add(["subtree"])
        export.export(wt, "target", format="dir")
        self.assertPathExists("target/subtree")
        # TODO(jelmer): Once iter_entries_by_dir supports nested tree iteration:
        # self.assertPathExists('target/subtree/file')

    def test_to_existing_empty_dir_success(self):
        self.build_tree(["source/", "source/a", "source/b/", "source/b/c"])
        wt = self.make_branch_and_tree("source")
        wt.add(["a", "b", "b/c"])
        wt.commit("1")
        self.build_tree(["target/"])
        export.export(wt, "target", format="dir")
        self.assertPathExists("target/a")
        self.assertPathExists("target/b")
        self.assertPathExists("target/b/c")

    def test_empty_subdir(self):
        self.build_tree(["source/", "source/a", "source/b/", "source/b/c"])
        wt = self.make_branch_and_tree("source")
        wt.add(["a", "b", "b/c"])
        wt.commit("1")
        self.build_tree(["target/"])
        export.export(wt, "target", format="dir", subdir="")
        self.assertPathExists("target/a")
        self.assertPathExists("target/b")
        self.assertPathExists("target/b/c")

    def test_to_existing_nonempty_dir_fail(self):
        self.build_tree(["source/", "source/a", "source/b/", "source/b/c"])
        wt = self.make_branch_and_tree("source")
        wt.add(["a", "b", "b/c"])
        wt.commit("1")
        self.build_tree(["target/", "target/foo"])
        self.assertRaises(errors.BzrError, export.export, wt, "target", format="dir")

    def test_existing_single_file(self):
        self.build_tree(["dir1/", "dir1/dir2/", "dir1/first", "dir1/dir2/second"])
        wtree = self.make_branch_and_tree("dir1")
        wtree.add(["dir2", "first", "dir2/second"])
        wtree.commit("1")
        export.export(wtree, "target1", format="dir", subdir="first")
        self.assertPathExists("target1/first")
        export.export(wtree, "target2", format="dir", subdir="dir2/second")
        self.assertPathExists("target2/second")

    def test_files_same_timestamp(self):
        builder = self.make_branch_builder("source")
        builder.start_series()
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"root-id", "directory", "")),
                ("add", ("a", b"a-id", "file", b"content\n")),
            ],
        )
        builder.build_snapshot(None, [("add", ("b", b"b-id", "file", b"content\n"))])
        builder.finish_series()
        b = builder.get_branch()
        b.lock_read()
        self.addCleanup(b.unlock)
        tree = b.basis_tree()
        orig_iter_files_bytes = tree.iter_files_bytes

        # Make iter_files_bytes slower, so we provoke mtime skew
        def iter_files_bytes(to_fetch):
            for thing in orig_iter_files_bytes(to_fetch):
                yield thing
                time.sleep(1)

        tree.iter_files_bytes = iter_files_bytes
        export.export(tree, "target", format="dir")
        t = self.get_transport("target")
        st_a = t.stat("a")
        st_b = t.stat("b")
        # All files must be given the same mtime.
        self.assertEqual(st_a.st_mtime, st_b.st_mtime)

    def test_files_per_file_timestamps(self):
        builder = self.make_branch_builder("source")
        builder.start_series()
        # Earliest allowable date on FAT32 filesystems is 1980-01-01
        a_time = time.mktime((1999, 12, 12, 0, 0, 0, 0, 0, 0))
        b_time = time.mktime((1980, 0o1, 0o1, 0, 0, 0, 0, 0, 0))
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"root-id", "directory", "")),
                ("add", ("a", b"a-id", "file", b"content\n")),
            ],
            timestamp=a_time,
        )
        builder.build_snapshot(
            None, [("add", ("b", b"b-id", "file", b"content\n"))], timestamp=b_time
        )
        builder.finish_series()
        b = builder.get_branch()
        b.lock_read()
        self.addCleanup(b.unlock)
        tree = b.basis_tree()
        export.export(tree, "target", format="dir", per_file_timestamps=True)
        t = self.get_transport("target")
        self.assertEqual(a_time, t.stat("a").st_mtime)
        self.assertEqual(b_time, t.stat("b").st_mtime)

    def test_subdir_files_per_timestamps(self):
        builder = self.make_branch_builder("source")
        builder.start_series()
        foo_time = time.mktime((1999, 12, 12, 0, 0, 0, 0, 0, 0))
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"root-id", "directory", "")),
                ("add", ("subdir", b"subdir-id", "directory", "")),
                ("add", ("subdir/foo.txt", b"foo-id", "file", b"content\n")),
            ],
            timestamp=foo_time,
        )
        builder.finish_series()
        b = builder.get_branch()
        b.lock_read()
        self.addCleanup(b.unlock)
        tree = b.basis_tree()
        export.export(
            tree, "target", format="dir", subdir="subdir", per_file_timestamps=True
        )
        t = self.get_transport("target")
        self.assertEqual(foo_time, t.stat("foo.txt").st_mtime)


class TarExporterTests(tests.TestCaseWithTransport):
    def test_xz(self):
        self.requireFeature(features.lzma)
        import lzma

        wt = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        wt.add(["a"])
        wt.commit("1")
        export.export(wt, "target.tar.xz", format="txz")
        tf = tarfile.open(fileobj=lzma.LZMAFile("target.tar.xz"))
        self.assertEqual(["target/a"], tf.getnames())

    def test_lzma(self):
        self.requireFeature(features.lzma)
        import lzma

        wt = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        wt.add(["a"])
        wt.commit("1")
        export.export(wt, "target.tar.lzma", format="tlzma")
        tf = tarfile.open(fileobj=lzma.LZMAFile("target.tar.lzma"))
        self.assertEqual(["target/a"], tf.getnames())

    def test_tgz(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        wt.add(["a"])
        wt.commit("1")
        export.export(wt, "target.tar.gz", format="tgz")
        tf = tarfile.open("target.tar.gz")
        self.assertEqual(["target/a"], tf.getnames())

    def test_tgz_consistent_mtime(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        wt.add(["a"])
        timestamp = 1547400500
        revid = wt.commit("1", timestamp=timestamp)
        revtree = wt.branch.repository.revision_tree(revid)
        export.export(revtree, "target.tar.gz", format="tgz")
        with gzip.GzipFile("target.tar.gz", "r") as f:
            f.read()
            self.assertEqual(int(f.mtime), timestamp)

    def test_tgz_ignores_dest_path(self):
        # The target path should not be a part of the target file.
        # (bug #102234)
        wt = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        wt.add(["a"])
        wt.commit("1")
        os.mkdir("testdir1")
        os.mkdir("testdir2")
        export.export(
            wt, "testdir1/target.tar.gz", format="tgz", per_file_timestamps=True
        )
        export.export(
            wt, "testdir2/target.tar.gz", format="tgz", per_file_timestamps=True
        )
        file1 = open("testdir1/target.tar.gz", "rb")
        self.addCleanup(file1.close)
        file2 = open("testdir1/target.tar.gz", "rb")
        self.addCleanup(file2.close)
        content1 = file1.read()
        content2 = file2.read()
        self.assertEqualDiff(content1, content2)
        # the gzip module doesn't have a way to read back to the original
        # filename, but it's stored as-is in the tarfile.
        self.assertFalse(b"testdir1" in content1)
        self.assertFalse(b"target.tar.gz" in content1)
        self.assertTrue(b"target.tar" in content1)

    def test_tbz2(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        wt.add(["a"])
        wt.commit("1")
        export.export(wt, "target.tar.bz2", format="tbz2")
        tf = tarfile.open("target.tar.bz2")
        self.assertEqual(["target/a"], tf.getnames())

    def test_export_tarball_generator(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        wt.add(["a"])
        wt.commit("1", timestamp=42)
        target = BytesIO()
        with wt.lock_read():
            target.writelines(tarball_generator(wt, "bar"))
        # Ball should now be closed.
        target.seek(0)
        ball2 = tarfile.open(None, "r", target)
        self.addCleanup(ball2.close)
        self.assertEqual(["bar/a"], ball2.getnames())


class ZipExporterTests(tests.TestCaseWithTransport):
    def test_per_file_timestamps(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("har", b"foo")])
        tree.add("har")
        # Earliest allowable date on FAT32 filesystems is 1980-01-01
        timestamp = 347151600
        tree.commit("setup", timestamp=timestamp)
        export.export(
            tree.basis_tree(), "test.zip", format="zip", per_file_timestamps=True
        )
        zfile = zipfile.ZipFile("test.zip")
        info = zfile.getinfo("test/har")
        self.assertEqual(time.localtime(timestamp)[:6], info.date_time)


class RootNameTests(tests.TestCase):
    def test_root_name(self):
        self.assertEqual("mytest", get_root_name("../mytest.tar"))
        self.assertEqual("mytar", get_root_name("mytar.tar"))
        self.assertEqual("mytar", get_root_name("mytar.tar.bz2"))
        self.assertEqual("tar.tar.tar", get_root_name("tar.tar.tar.tgz"))
        self.assertEqual("bzr-0.0.5", get_root_name("bzr-0.0.5.tar.gz"))
        self.assertEqual("bzr-0.0.5", get_root_name("bzr-0.0.5.zip"))
        self.assertEqual("bzr-0.0.5", get_root_name("bzr-0.0.5"))
        self.assertEqual("mytar", get_root_name("a/long/path/mytar.tgz"))
        self.assertEqual("other", get_root_name("../parent/../dir/other.tbz2"))
        self.assertEqual("", get_root_name("-"))



# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests for breezy.git.tests."""

import stat

from dulwich.objects import Blob

from breezy.bzr.inventorytree import InventoryTreeChange as TreeChange
from breezy.delta import TreeDelta
from breezy.errors import PathsNotVersionedError
from breezy.git.mapping import default_mapping
from breezy.git.tree import changes_from_git_changes, tree_delta_from_git_changes
from breezy.tests import TestCase, TestCaseWithTransport
from breezy.workingtree import WorkingTree

REG_MODE = stat.S_IFREG | 0o644


class ChangesFromGitChangesTests(TestCase):
    def setUp(self):
        super().setUp()
        self.maxDiff = None
        self.mapping = default_mapping

    def transform(
        self,
        changes,
        specific_files=None,
        include_unchanged=False,
        source_extras=None,
        target_extras=None,
    ):
        return list(
            changes_from_git_changes(
                changes,
                self.mapping,
                specific_files=specific_files,
                include_unchanged=include_unchanged,
                source_extras=source_extras,
                target_extras=target_extras,
            )
        )

    def test_empty(self):
        self.assertEqual([], self.transform([]))

    def test_modified(self):
        a = Blob.from_string(b"a")
        b = Blob.from_string(b"b")
        self.assertEqual(
            [
                TreeChange(
                    b"git:a",
                    ("a", "a"),
                    True,
                    (True, True),
                    (b"TREE_ROOT", b"TREE_ROOT"),
                    ("a", "a"),
                    ("file", "file"),
                    (False, False),
                    False,
                )
            ],
            self.transform(
                [
                    (
                        "modify",
                        (b"a", stat.S_IFREG | 0o644, a),
                        (b"a", stat.S_IFREG | 0o644, b),
                    )
                ]
            ),
        )

    def test_kind_changed(self):
        a = Blob.from_string(b"a")
        b = Blob.from_string(b"target")
        self.assertEqual(
            [
                TreeChange(
                    b"git:a",
                    ("a", "a"),
                    True,
                    (True, True),
                    (b"TREE_ROOT", b"TREE_ROOT"),
                    ("a", "a"),
                    ("file", "symlink"),
                    (False, False),
                    False,
                )
            ],
            self.transform(
                [("modify", (b"a", stat.S_IFREG | 0o644, a), (b"a", stat.S_IFLNK, b))]
            ),
        )

    def test_rename_no_changes(self):
        a = Blob.from_string(b"a")
        self.assertEqual(
            [
                TreeChange(
                    b"git:old",
                    ("old", "a"),
                    False,
                    (True, True),
                    (b"TREE_ROOT", b"TREE_ROOT"),
                    ("old", "a"),
                    ("file", "file"),
                    (False, False),
                    False,
                )
            ],
            self.transform(
                [
                    (
                        "rename",
                        (b"old", stat.S_IFREG | 0o644, a),
                        (b"a", stat.S_IFREG | 0o644, a),
                    )
                ]
            ),
        )

    def test_rename_and_modify(self):
        a = Blob.from_string(b"a")
        b = Blob.from_string(b"b")
        self.assertEqual(
            [
                TreeChange(
                    b"git:a",
                    ("a", "b"),
                    True,
                    (True, True),
                    (b"TREE_ROOT", b"TREE_ROOT"),
                    ("a", "b"),
                    ("file", "file"),
                    (False, False),
                    False,
                )
            ],
            self.transform(
                [
                    (
                        "rename",
                        (b"a", stat.S_IFREG | 0o644, a),
                        (b"b", stat.S_IFREG | 0o644, b),
                    )
                ]
            ),
        )

    def test_copy_no_changes(self):
        a = Blob.from_string(b"a")
        self.assertEqual(
            [
                TreeChange(
                    b"git:a",
                    ("old", "a"),
                    False,
                    (True, True),
                    (b"TREE_ROOT", b"TREE_ROOT"),
                    ("old", "a"),
                    ("file", "file"),
                    (False, False),
                    True,
                )
            ],
            self.transform(
                [
                    (
                        "copy",
                        (b"old", stat.S_IFREG | 0o644, a),
                        (b"a", stat.S_IFREG | 0o644, a),
                    )
                ]
            ),
        )

    def test_copy_and_modify(self):
        a = Blob.from_string(b"a")
        b = Blob.from_string(b"b")
        self.assertEqual(
            [
                TreeChange(
                    b"git:b",
                    ("a", "b"),
                    True,
                    (True, True),
                    (b"TREE_ROOT", b"TREE_ROOT"),
                    ("a", "b"),
                    ("file", "file"),
                    (False, False),
                    True,
                )
            ],
            self.transform(
                [
                    (
                        "copy",
                        (b"a", stat.S_IFREG | 0o644, a),
                        (b"b", stat.S_IFREG | 0o644, b),
                    )
                ]
            ),
        )

    def test_add(self):
        b = Blob.from_string(b"b")
        self.assertEqual(
            [
                TreeChange(
                    b"git:a",
                    (None, "a"),
                    True,
                    (False, True),
                    (None, b"TREE_ROOT"),
                    (None, "a"),
                    (None, "file"),
                    (None, False),
                    False,
                )
            ],
            self.transform(
                [("add", (None, None, None), (b"a", stat.S_IFREG | 0o644, b))]
            ),
        )

    def test_delete(self):
        b = Blob.from_string(b"b")
        self.assertEqual(
            [
                TreeChange(
                    b"git:a",
                    ("a", None),
                    True,
                    (True, False),
                    (b"TREE_ROOT", None),
                    ("a", None),
                    ("file", None),
                    (False, None),
                    False,
                )
            ],
            self.transform(
                [("remove", (b"a", stat.S_IFREG | 0o644, b), (None, None, None))]
            ),
        )

    def test_unchanged(self):
        b = Blob.from_string(b"b")
        self.assertEqual(
            [
                TreeChange(
                    b"git:a",
                    ("a", "a"),
                    False,
                    (True, True),
                    (b"TREE_ROOT", b"TREE_ROOT"),
                    ("a", "a"),
                    ("file", "file"),
                    (False, False),
                    False,
                )
            ],
            self.transform(
                [
                    (
                        "unchanged",
                        (b"a", stat.S_IFREG | 0o644, b),
                        (b"a", stat.S_IFREG | 0o644, b),
                    )
                ],
                include_unchanged=True,
            ),
        )
        self.assertEqual(
            [],
            self.transform(
                [
                    (
                        "unchanged",
                        (b"a", stat.S_IFREG | 0o644, b),
                        (b"a", stat.S_IFREG | 0o644, b),
                    )
                ],
                include_unchanged=False,
            ),
        )

    def test_unversioned(self):
        b = Blob.from_string(b"b")
        self.assertEqual(
            [
                TreeChange(
                    None,
                    (None, "a"),
                    True,
                    (False, False),
                    (None, b"TREE_ROOT"),
                    (None, "a"),
                    (None, "file"),
                    (None, False),
                    False,
                )
            ],
            self.transform(
                [("add", (None, None, None), (b"a", stat.S_IFREG | 0o644, b))],
                target_extras={b"a"},
            ),
        )
        self.assertEqual(
            [
                TreeChange(
                    None,
                    ("a", "a"),
                    False,
                    (False, False),
                    (b"TREE_ROOT", b"TREE_ROOT"),
                    ("a", "a"),
                    ("file", "file"),
                    (False, False),
                    False,
                )
            ],
            self.transform(
                [
                    (
                        "add",
                        (b"a", stat.S_IFREG | 0o644, b),
                        (b"a", stat.S_IFREG | 0o644, b),
                    )
                ],
                source_extras={b"a"},
                target_extras={b"a"},
            ),
        )


class DeltaFromGitChangesTests(TestCase):
    def setUp(self):
        super().setUp()
        self.maxDiff = None
        self.mapping = default_mapping

    def transform(
        self,
        changes,
        specific_files=None,
        require_versioned=False,
        include_root=False,
        source_extras=None,
        target_extras=None,
    ):
        return tree_delta_from_git_changes(
            changes,
            (self.mapping, self.mapping),
            specific_files=specific_files,
            require_versioned=require_versioned,
            include_root=include_root,
            source_extras=source_extras,
            target_extras=target_extras,
        )

    def test_empty(self):
        self.assertEqual(TreeDelta(), self.transform([]))

    def test_modified(self):
        a = Blob.from_string(b"a")
        b = Blob.from_string(b"b")
        delta = self.transform(
            [
                (
                    "modify",
                    (b"a", stat.S_IFREG | 0o644, a),
                    (b"a", stat.S_IFREG | 0o644, b),
                )
            ]
        )
        expected_delta = TreeDelta()
        expected_delta.modified.append(
            TreeChange(
                b"git:a",
                ("a", "a"),
                True,
                (True, True),
                (b"TREE_ROOT", b"TREE_ROOT"),
                ("a", "a"),
                ("file", "file"),
                (False, False),
                False,
            )
        )
        self.assertEqual(expected_delta, delta)

    def test_rename_no_changes(self):
        a = Blob.from_string(b"a")
        delta = self.transform(
            [
                (
                    "rename",
                    (b"old", stat.S_IFREG | 0o644, a),
                    (b"a", stat.S_IFREG | 0o644, a),
                )
            ]
        )
        expected_delta = TreeDelta()
        expected_delta.renamed.append(
            TreeChange(
                b"git:old",
                ("old", "a"),
                False,
                (True, True),
                (b"TREE_ROOT", b"TREE_ROOT"),
                ("old", "a"),
                ("file", "file"),
                (False, False),
                False,
            )
        )
        self.assertEqual(expected_delta, delta)

    def test_rename_and_modify(self):
        a = Blob.from_string(b"a")
        b = Blob.from_string(b"b")
        delta = self.transform(
            [
                (
                    "rename",
                    (b"a", stat.S_IFREG | 0o644, a),
                    (b"b", stat.S_IFREG | 0o644, b),
                )
            ]
        )
        expected_delta = TreeDelta()
        expected_delta.renamed.append(
            TreeChange(
                b"git:a",
                ("a", "b"),
                True,
                (True, True),
                (b"TREE_ROOT", b"TREE_ROOT"),
                ("a", "b"),
                ("file", "file"),
                (False, False),
                False,
            )
        )
        self.assertEqual(delta, expected_delta)

    def test_copy_no_changes(self):
        a = Blob.from_string(b"a")
        delta = self.transform(
            [
                (
                    "copy",
                    (b"old", stat.S_IFREG | 0o644, a),
                    (b"a", stat.S_IFREG | 0o644, a),
                )
            ]
        )
        expected_delta = TreeDelta()
        expected_delta.copied.append(
            TreeChange(
                b"git:a",
                ("old", "a"),
                False,
                (True, True),
                (b"TREE_ROOT", b"TREE_ROOT"),
                ("old", "a"),
                ("file", "file"),
                (False, False),
                True,
            )
        )
        self.assertEqual(expected_delta, delta)

    def test_copy_and_modify(self):
        a = Blob.from_string(b"a")
        b = Blob.from_string(b"b")
        delta = self.transform(
            [("copy", (b"a", stat.S_IFREG | 0o644, a), (b"b", stat.S_IFREG | 0o644, b))]
        )
        expected_delta = TreeDelta()
        expected_delta.copied.append(
            TreeChange(
                b"git:b",
                ("a", "b"),
                True,
                (True, True),
                (b"TREE_ROOT", b"TREE_ROOT"),
                ("a", "b"),
                ("file", "file"),
                (False, False),
                True,
            )
        )
        self.assertEqual(expected_delta, delta)

    def test_add(self):
        b = Blob.from_string(b"b")
        delta = self.transform(
            [("add", (None, None, None), (b"a", stat.S_IFREG | 0o644, b))]
        )
        expected_delta = TreeDelta()
        expected_delta.added.append(
            TreeChange(
                b"git:a",
                (None, "a"),
                True,
                (False, True),
                (None, b"TREE_ROOT"),
                (None, "a"),
                (None, "file"),
                (None, False),
                False,
            )
        )
        self.assertEqual(delta, expected_delta)

    def test_delete(self):
        b = Blob.from_string(b"b")
        delta = self.transform(
            [("remove", (b"a", stat.S_IFREG | 0o644, b), (None, None, None))]
        )
        expected_delta = TreeDelta()
        expected_delta.removed.append(
            TreeChange(
                b"git:a",
                ("a", None),
                True,
                (True, False),
                (b"TREE_ROOT", None),
                ("a", None),
                ("file", None),
                (False, None),
                False,
            )
        )
        self.assertEqual(delta, expected_delta)

    def test_unchanged(self):
        b = Blob.from_string(b"b")
        self.transform(
            [
                (
                    "unchanged",
                    (b"a", stat.S_IFREG | 0o644, b),
                    (b"a", stat.S_IFREG | 0o644, b),
                )
            ]
        )
        expected_delta = TreeDelta()
        expected_delta.unchanged.append(
            TreeChange(
                b"git:a",
                ("a", "a"),
                False,
                (True, True),
                (b"TREE_ROOT", b"TREE_ROOT"),
                ("a", "a"),
                ("file", "file"),
                (False, False),
                False,
            )
        )

    def test_unversioned(self):
        b = Blob.from_string(b"b")
        delta = self.transform(
            [("add", (None, None, None), (b"a", stat.S_IFREG | 0o644, b))],
            target_extras={b"a"},
        )
        expected_delta = TreeDelta()
        expected_delta.unversioned.append(
            TreeChange(
                None,
                (None, "a"),
                True,
                (False, False),
                (None, b"TREE_ROOT"),
                (None, "a"),
                (None, "file"),
                (None, False),
                False,
            )
        )
        self.assertEqual(delta, expected_delta)
        delta = self.transform(
            [("add", (b"a", stat.S_IFREG | 0o644, b), (b"a", stat.S_IFREG | 0o644, b))],
            source_extras={b"a"},
            target_extras={b"a"},
        )
        expected_delta = TreeDelta()
        expected_delta.unversioned.append(
            TreeChange(
                None,
                ("a", "a"),
                False,
                (False, False),
                (b"TREE_ROOT", b"TREE_ROOT"),
                ("a", "a"),
                ("file", "file"),
                (False, False),
                False,
            )
        )
        self.assertEqual(delta, expected_delta)

    def test_kind_change(self):
        a = Blob.from_string(b"a")
        b = Blob.from_string(b"target")
        delta = self.transform(
            [("modify", (b"a", stat.S_IFREG | 0o644, a), (b"a", stat.S_IFLNK, b))]
        )
        expected_delta = TreeDelta()
        expected_delta.kind_changed.append(
            TreeChange(
                b"git:a",
                ("a", "a"),
                True,
                (True, True),
                (b"TREE_ROOT", b"TREE_ROOT"),
                ("a", "a"),
                ("file", "symlink"),
                (False, False),
                False,
            )
        )
        self.assertEqual(expected_delta, delta)


class FindRelatedPathsAcrossTrees(TestCaseWithTransport):
    def test_none(self):
        self.make_branch_and_tree("t1", format="git")
        wt = WorkingTree.open("t1")
        self.assertIs(None, wt.find_related_paths_across_trees(None))

    def test_empty(self):
        self.make_branch_and_tree("t1", format="git")
        wt = WorkingTree.open("t1")
        self.assertEqual([], list(wt.find_related_paths_across_trees([])))

    def test_directory(self):
        self.make_branch_and_tree("t1", format="git")
        wt = WorkingTree.open("t1")
        self.build_tree(["t1/dir/", "t1/dir/file"])
        wt.add(["dir", "dir/file"])
        self.assertEqual(
            ["dir/file"], list(wt.find_related_paths_across_trees(["dir/file"]))
        )
        self.assertEqual(["dir"], list(wt.find_related_paths_across_trees(["dir"])))

    def test_empty_directory(self):
        self.make_branch_and_tree("t1", format="git")
        wt = WorkingTree.open("t1")
        self.build_tree(["t1/dir/"])
        wt.add(["dir"])
        self.assertEqual(["dir"], list(wt.find_related_paths_across_trees(["dir"])))
        self.assertRaises(
            PathsNotVersionedError, wt.find_related_paths_across_trees, ["dir/file"]
        )

    def test_missing(self):
        self.make_branch_and_tree("t1", format="git")
        wt = WorkingTree.open("t1")
        self.assertRaises(
            PathsNotVersionedError, wt.find_related_paths_across_trees, ["file"]
        )

    def test_not_versioned(self):
        self.make_branch_and_tree("t1", format="git")
        self.make_branch_and_tree("t2", format="git")
        wt1 = WorkingTree.open("t1")
        wt2 = WorkingTree.open("t2")
        self.build_tree(["t1/file"])
        self.build_tree(["t2/file"])
        self.assertRaises(
            PathsNotVersionedError, wt1.find_related_paths_across_trees, ["file"], [wt2]
        )

    def test_single(self):
        self.make_branch_and_tree("t1", format="git")
        wt = WorkingTree.open("t1")
        self.build_tree(["t1/file"])
        wt.add("file")
        self.assertEqual(["file"], list(wt.find_related_paths_across_trees(["file"])))



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
from io import BytesIO

from breezy.tests.matchers import MatchesTreeChanges
from breezy.tests.per_tree import TestCaseWithTree

from ... import revision as _mod_revision
from ... import tests, trace
from ...diff import show_diff_trees
from ...merge import Merge3Merger, Merger
from ...transform import ROOT_PARENT, resolve_conflicts
from ...tree import TreeChange, find_previous_path
from ..features import SymlinkFeature, UnicodeFilenameFeature


class TestTransformPreview(TestCaseWithTree):
    def create_tree(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("a", b"content 1")])
        tree.add("a")
        revid1 = tree.commit("rev1")
        return tree.branch.repository.revision_tree(revid1)

    def get_empty_preview(self):
        repository = self.make_repository("repo")
        tree = repository.revision_tree(_mod_revision.NULL_REVISION)
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        return preview

    def test_transform_preview(self):
        revision_tree = self.create_tree()
        preview = revision_tree.preview_transform()
        self.addCleanup(preview.finalize)

    def test_transform_preview_tree(self):
        revision_tree = self.create_tree()
        preview = revision_tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.get_preview_tree()

    def test_transform_new_file(self):
        revision_tree = self.create_tree()
        preview = revision_tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.new_file("file2", preview.root, [b"content B\n"], b"file2-id")
        preview_tree = preview.get_preview_tree()
        self.assertEqual(preview_tree.kind("file2"), "file")
        with preview_tree.get_file("file2") as f:
            self.assertEqual(f.read(), b"content B\n")

    def test_diff_preview_tree(self):
        revision_tree = self.create_tree()
        preview = revision_tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.new_file("file2", preview.root, [b"content B\n"], b"file2-id")
        preview_tree = preview.get_preview_tree()
        out = BytesIO()
        show_diff_trees(revision_tree, preview_tree, out)
        lines = out.getvalue().splitlines()
        self.assertEqual(lines[0], b"=== added file 'file2'")
        # 3 lines of diff administrivia
        self.assertEqual(lines[4], b"+content B")

    def test_unsupported_symlink_diff(self):
        self.requireFeature(SymlinkFeature(self.test_dir))
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("a", "content 1")])
        tree.add("a")
        os.symlink("a", "foo")
        tree.add("foo")
        revid1 = tree.commit("rev1")
        revision_tree = tree.branch.repository.revision_tree(revid1)
        os_symlink = getattr(os, "symlink", None)
        os.symlink = None
        try:
            preview = revision_tree.preview_transform()
            self.addCleanup(preview.finalize)
            preview.delete_versioned(preview.trans_id_tree_path("foo"))
            preview_tree = preview.get_preview_tree()
            out = BytesIO()
            log = BytesIO()
            trace.push_log_file(log)
            show_diff_trees(revision_tree, preview_tree, out)
            out.getvalue().splitlines()
        finally:
            os.symlink = os_symlink
        self.assertContainsRe(
            log.getvalue(),
            b'Ignoring "foo" as symlinks are not supported on this filesystem',
        )

    def test_transform_conflicts(self):
        revision_tree = self.create_tree()
        preview = revision_tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.new_file("a", preview.root, [b"content 2"])
        resolve_conflicts(preview)
        trans_id = preview.trans_id_tree_path("a")
        self.assertEqual("a.moved", preview.final_name(trans_id))

    def get_tree_and_preview_tree(self):
        revision_tree = self.create_tree()
        preview = revision_tree.preview_transform()
        self.addCleanup(preview.finalize)
        a_trans_id = preview.trans_id_tree_path("a")
        preview.delete_contents(a_trans_id)
        preview.create_file([b"b content"], a_trans_id)
        preview_tree = preview.get_preview_tree()
        return revision_tree, preview_tree

    def test_iter_changes(self):
        revision_tree, preview_tree = self.get_tree_and_preview_tree()
        self.assertThat(
            preview_tree.iter_changes(revision_tree),
            MatchesTreeChanges(
                revision_tree,
                preview_tree,
                [
                    (
                        ("a", "a"),
                        True,
                        (True, True),
                        ("a", "a"),
                        ("file", "file"),
                        (False, False),
                        False,
                    )
                ],
            ),
        )

    def test_include_unchanged_succeeds(self):
        revision_tree, preview_tree = self.get_tree_and_preview_tree()
        changes = preview_tree.iter_changes(revision_tree, include_unchanged=True)

        root_entry = TreeChange(
            ("", ""),
            False,
            (True, True),
            ("", ""),
            ("directory", "directory"),
            (False, False),
            False,
        )
        a_entry = TreeChange(
            ("a", "a"),
            True,
            (True, True),
            ("a", "a"),
            ("file", "file"),
            (False, False),
            False,
        )

        self.assertThat(
            changes,
            MatchesTreeChanges(revision_tree, preview_tree, [root_entry, a_entry]),
        )

    def test_specific_files(self):
        revision_tree, preview_tree = self.get_tree_and_preview_tree()
        changes = preview_tree.iter_changes(revision_tree, specific_files=[""])
        a_entry = (
            ("a", "a"),
            True,
            (True, True),
            ("a", "a"),
            ("file", "file"),
            (False, False),
            False,
        )

        self.assertThat(
            changes, MatchesTreeChanges(revision_tree, preview_tree, [a_entry])
        )

    def test_want_unversioned(self):
        revision_tree, preview_tree = self.get_tree_and_preview_tree()
        changes = preview_tree.iter_changes(revision_tree, want_unversioned=True)
        a_entry = TreeChange(
            ("a", "a"),
            True,
            (True, True),
            ("a", "a"),
            ("file", "file"),
            (False, False),
            False,
        )

        self.assertThat(
            changes, MatchesTreeChanges(revision_tree, preview_tree, [a_entry])
        )

    def test_ignore_extra_trees_no_specific_files(self):
        # extra_trees is harmless without specific_files, so we'll silently
        # accept it, even though we won't use it.
        revision_tree, preview_tree = self.get_tree_and_preview_tree()
        preview_tree.iter_changes(revision_tree, extra_trees=[preview_tree])

    def test_ignore_require_versioned_no_specific_files(self):
        # require_versioned is meaningless without specific_files.
        revision_tree, preview_tree = self.get_tree_and_preview_tree()
        preview_tree.iter_changes(revision_tree, require_versioned=False)

    def test_ignore_pb(self):
        # pb could be supported, but TT.iter_changes doesn't support it.
        revision_tree, preview_tree = self.get_tree_and_preview_tree()
        preview_tree.iter_changes(revision_tree)

    def test_kind(self):
        revision_tree = self.create_tree()
        preview = revision_tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.new_file("file", preview.root, [b"contents"], b"file-id")
        preview.new_directory("directory", preview.root, b"dir-id")
        preview_tree = preview.get_preview_tree()
        self.assertEqual("file", preview_tree.kind("file"))
        self.assertEqual("directory", preview_tree.kind("directory"))

    def test_get_file_mtime(self):
        preview = self.get_empty_preview()
        file_trans_id = preview.new_file(
            "file", preview.root, [b"contents"], b"file-id"
        )
        limbo_path = preview._limbo_name(file_trans_id)
        preview_tree = preview.get_preview_tree()
        self.assertEqual(
            os.stat(limbo_path).st_mtime, preview_tree.get_file_mtime("file")
        )

    def test_get_file_mtime_renamed(self):
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/file"])
        work_tree.add("file")
        preview = work_tree.preview_transform()
        self.addCleanup(preview.finalize)
        file_trans_id = preview.trans_id_tree_path("file")
        preview.adjust_path("renamed", preview.root, file_trans_id)
        preview_tree = preview.get_preview_tree()
        preview_tree.get_file_mtime("renamed")
        work_tree.get_file_mtime("file")

    def test_get_file_size(self):
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/old", b"old")])
        work_tree.add("old")
        preview = work_tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.new_file("name", preview.root, [b"contents"], b"new-id", "executable")
        tree = preview.get_preview_tree()
        self.assertEqual(len("old"), tree.get_file_size("old"))
        self.assertEqual(len("contents"), tree.get_file_size("name"))

    def test_get_file(self):
        preview = self.get_empty_preview()
        preview.new_file("file", preview.root, [b"contents"], b"file-id")
        preview_tree = preview.get_preview_tree()
        with preview_tree.get_file("file") as tree_file:
            self.assertEqual(b"contents", tree_file.read())

    def test_get_symlink_target(self):
        self.requireFeature(SymlinkFeature(self.test_dir))
        preview = self.get_empty_preview()
        preview.new_symlink("symlink", preview.root, "target", b"symlink-id")
        preview_tree = preview.get_preview_tree()
        self.assertEqual("target", preview_tree.get_symlink_target("symlink"))

    def test_all_file_ids(self):
        if not self.workingtree_format.supports_setting_file_ids:
            raise tests.TestNotApplicable("format does not support setting file ids")
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/b", "tree/c"])
        tree.add(["a", "b", "c"], ids=[b"a-id", b"b-id", b"c-id"])
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.unversion_file(preview.trans_id_file_id(b"b-id"))
        c_trans_id = preview.trans_id_file_id(b"c-id")
        preview.unversion_file(c_trans_id)
        preview.version_file(c_trans_id, file_id=b"c-id")
        preview_tree = preview.get_preview_tree()
        self.assertEqual(
            {b"a-id", b"c-id", tree.path2id("")}, preview_tree.all_file_ids()
        )

    def test_path2id_deleted_unchanged(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/unchanged", "tree/deleted"])
        tree.add(["unchanged", "deleted"])
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.unversion_file(preview.trans_id_tree_path("deleted"))
        preview_tree = preview.get_preview_tree()
        self.assertEqual(
            "unchanged", find_previous_path(preview_tree, tree, "unchanged")
        )
        self.assertFalse(preview_tree.is_versioned("deleted"))

    def test_path2id_created(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/unchanged"])
        tree.add(["unchanged"])
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.new_file(
            "new", preview.trans_id_tree_path("unchanged"), [b"contents"], b"new-id"
        )
        preview_tree = preview.get_preview_tree()
        self.assertTrue(preview_tree.is_versioned("unchanged/new"))
        if self.workingtree_format.supports_setting_file_ids:
            self.assertEqual(b"new-id", preview_tree.path2id("unchanged/new"))

    def test_path2id_moved(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/old_parent/", "tree/old_parent/child"])
        tree.add(["old_parent", "old_parent/child"])
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        new_parent = preview.new_directory("new_parent", preview.root, b"new_parent-id")
        preview.adjust_path(
            "child", new_parent, preview.trans_id_tree_path("old_parent/child")
        )
        preview_tree = preview.get_preview_tree()
        self.assertFalse(preview_tree.is_versioned("old_parent/child"))
        self.assertEqual(
            "new_parent/child",
            find_previous_path(tree, preview_tree, "old_parent/child"),
        )
        if self.workingtree_format.supports_setting_file_ids:
            self.assertEqual(
                tree.path2id("old_parent/child"),
                preview_tree.path2id("new_parent/child"),
            )

    def test_path2id_renamed_parent(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/old_name/", "tree/old_name/child"])
        tree.add(["old_name", "old_name/child"])
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.adjust_path(
            "new_name", preview.root, preview.trans_id_tree_path("old_name")
        )
        preview_tree = preview.get_preview_tree()
        self.assertFalse(preview_tree.is_versioned("old_name/child"))
        self.assertEqual(
            "new_name/child", find_previous_path(tree, preview_tree, "old_name/child")
        )
        if tree.supports_setting_file_ids():
            self.assertEqual(
                tree.path2id("old_name/child"), preview_tree.path2id("new_name/child")
            )

    def assertMatchingIterEntries(self, tt, specific_files=None):
        preview_tree = tt.get_preview_tree()
        preview_result = list(
            preview_tree.iter_entries_by_dir(specific_files=specific_files)
        )
        tree = tt._tree
        tt.apply()
        actual_result = list(tree.iter_entries_by_dir(specific_files=specific_files))
        self.assertEqual(actual_result, preview_result)

    def test_iter_entries_by_dir_new(self):
        tree = self.make_branch_and_tree("tree")
        tt = tree.transform()
        tt.new_file("new", tt.root, [b"contents"], b"new-id")
        self.assertMatchingIterEntries(tt)

    def test_iter_entries_by_dir_deleted(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/deleted"])
        tree.add("deleted")
        tt = tree.transform()
        tt.delete_contents(tt.trans_id_tree_path("deleted"))
        self.assertMatchingIterEntries(tt)

    def test_iter_entries_by_dir_unversioned(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/removed"])
        tree.add("removed")
        tt = tree.transform()
        tt.unversion_file(tt.trans_id_tree_path("removed"))
        self.assertMatchingIterEntries(tt)

    def test_iter_entries_by_dir_moved(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/moved", "tree/new_parent/"])
        tree.add(["moved", "new_parent"])
        tt = tree.transform()
        tt.adjust_path(
            "moved", tt.trans_id_tree_path("new_parent"), tt.trans_id_tree_path("moved")
        )
        self.assertMatchingIterEntries(tt)

    def test_iter_entries_by_dir_specific_files(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/parent/", "tree/parent/child"])
        tree.add(["parent", "parent/child"])
        tt = tree.transform()
        self.assertMatchingIterEntries(tt, ["", "parent/child"])

    def test_symlink_content_summary(self):
        self.requireFeature(SymlinkFeature(self.test_dir))
        preview = self.get_empty_preview()
        preview.new_symlink("path", preview.root, "target", b"path-id")
        summary = preview.get_preview_tree().path_content_summary("path")
        self.assertEqual(("symlink", None, None, "target"), summary)

    def test_missing_content_summary(self):
        preview = self.get_empty_preview()
        summary = preview.get_preview_tree().path_content_summary("path")
        self.assertEqual(("missing", None, None, None), summary)

    def test_deleted_content_summary(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/path/"])
        tree.add("path")
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.delete_contents(preview.trans_id_tree_path("path"))
        summary = preview.get_preview_tree().path_content_summary("path")
        self.assertEqual(("missing", None, None, None), summary)

    def test_file_content_summary_executable(self):
        preview = self.get_empty_preview()
        path_id = preview.new_file("path", preview.root, [b"contents"], b"path-id")
        preview.set_executability(True, path_id)
        summary = preview.get_preview_tree().path_content_summary("path")
        self.assertEqual(4, len(summary))
        self.assertEqual("file", summary[0])
        # size must be known
        self.assertEqual(len("contents"), summary[1])
        # executable
        self.assertEqual(True, summary[2])
        # will not have hash (not cheap to determine)
        self.assertIs(None, summary[3])

    def test_change_executability(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/path"])
        tree.add("path")
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        path_id = preview.trans_id_tree_path("path")
        preview.set_executability(True, path_id)
        summary = preview.get_preview_tree().path_content_summary("path")
        self.assertEqual(True, summary[2])

    def test_file_content_summary_non_exec(self):
        preview = self.get_empty_preview()
        preview.new_file("path", preview.root, [b"contents"], b"path-id")
        summary = preview.get_preview_tree().path_content_summary("path")
        self.assertEqual(4, len(summary))
        self.assertEqual("file", summary[0])
        # size must be known
        self.assertEqual(len("contents"), summary[1])
        # not executable
        self.assertEqual(False, summary[2])
        # will not have hash (not cheap to determine)
        self.assertIs(None, summary[3])

    def test_dir_content_summary(self):
        preview = self.get_empty_preview()
        preview.new_directory("path", preview.root, b"path-id")
        summary = preview.get_preview_tree().path_content_summary("path")
        self.assertEqual(("directory", None, None, None), summary)

    def test_tree_content_summary(self):
        preview = self.get_empty_preview()
        path = preview.new_directory("path", preview.root, b"path-id")
        preview.set_tree_reference(b"rev-1", path)
        summary = preview.get_preview_tree().path_content_summary("path")
        self.assertEqual(4, len(summary))
        self.assertEqual("tree-reference", summary[0])

    def test_annotate(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", b"a\n")])
        tree.add("file")
        revid1 = tree.commit("a")
        self.build_tree_contents([("tree/file", b"a\nb\n")])
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        file_trans_id = preview.trans_id_tree_path("file")
        preview.delete_contents(file_trans_id)
        preview.create_file([b"a\nb\nc\n"], file_trans_id)
        preview_tree = preview.get_preview_tree()
        expected = [
            (revid1, b"a\n"),
            (b"me:", b"b\n"),
            (b"me:", b"c\n"),
        ]
        annotation = preview_tree.annotate_iter("file", default_revision=b"me:")
        self.assertEqual(expected, annotation)

    def test_annotate_missing(self):
        preview = self.get_empty_preview()
        preview.new_file("file", preview.root, [b"a\nb\nc\n"], b"file-id")
        preview_tree = preview.get_preview_tree()
        expected = [
            (b"me:", b"a\n"),
            (b"me:", b"b\n"),
            (b"me:", b"c\n"),
        ]
        annotation = preview_tree.annotate_iter("file", default_revision=b"me:")
        self.assertEqual(expected, annotation)

    def test_annotate_rename(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", b"a\n")])
        tree.add("file")
        revid1 = tree.commit("a")
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        file_trans_id = preview.trans_id_tree_path("file")
        preview.adjust_path("newname", preview.root, file_trans_id)
        preview_tree = preview.get_preview_tree()
        expected = [
            (revid1, b"a\n"),
        ]
        annotation = preview_tree.annotate_iter("newname", default_revision=b"me:")
        self.assertEqual(expected, annotation)
        annotation = preview_tree.annotate_iter("file", default_revision=b"me:")
        self.assertIs(None, annotation)

    def test_annotate_deleted(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", b"a\n")])
        tree.add("file")
        tree.commit("a")
        self.build_tree_contents([("tree/file", b"a\nb\n")])
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        file_trans_id = preview.trans_id_tree_path("file")
        preview.delete_contents(file_trans_id)
        preview_tree = preview.get_preview_tree()
        annotation = preview_tree.annotate_iter("file", default_revision=b"me:")
        self.assertIs(None, annotation)

    def test_stored_kind(self):
        preview = self.get_empty_preview()
        preview.new_file("file", preview.root, [b"a\nb\nc\n"], b"file-id")
        preview_tree = preview.get_preview_tree()
        self.assertEqual("file", preview_tree.stored_kind("file"))

    def test_is_executable(self):
        preview = self.get_empty_preview()
        trans_id = preview.new_file("file", preview.root, [b"a\nb\nc\n"], b"file-id")
        preview.set_executability(True, trans_id)
        preview_tree = preview.get_preview_tree()
        self.assertEqual(True, preview_tree.is_executable("file"))

    def test_get_set_parent_ids(self):
        revision_tree, preview_tree = self.get_tree_and_preview_tree()
        self.assertEqual([], preview_tree.get_parent_ids())
        preview_tree.set_parent_ids([revision_tree.get_revision_id()])
        self.assertEqual(
            [revision_tree.get_revision_id()], preview_tree.get_parent_ids()
        )

    def test_plan_file_merge(self):
        work_a = self.make_branch_and_tree("wta")
        self.build_tree_contents([("wta/file", b"a\nb\nc\nd\n")])
        work_a.add("file")
        base_id = work_a.commit("base version")
        tree_b = work_a.controldir.sprout("wtb").open_workingtree()
        preview = work_a.preview_transform()
        self.addCleanup(preview.finalize)
        trans_id = preview.trans_id_tree_path("file")
        preview.delete_contents(trans_id)
        preview.create_file([b"b\nc\nd\ne\n"], trans_id)
        self.build_tree_contents([("wtb/file", b"a\nc\nd\nf\n")])
        tree_a = preview.get_preview_tree()
        if not getattr(tree_a, "plan_file_merge", None):
            self.skipTest("tree does not support file merge planning")
        tree_a.set_parent_ids([base_id])
        self.addCleanup(tree_b.lock_read().unlock)
        self.assertEqual(
            [
                ("killed-a", b"a\n"),
                ("killed-b", b"b\n"),
                ("unchanged", b"c\n"),
                ("unchanged", b"d\n"),
                ("new-a", b"e\n"),
                ("new-b", b"f\n"),
            ],
            list(tree_a.plan_file_merge("file", tree_b)),
        )

    def test_plan_file_merge_revision_tree(self):
        work_a = self.make_branch_and_tree("wta")
        self.build_tree_contents([("wta/file", b"a\nb\nc\nd\n")])
        work_a.add("file")
        base_id = work_a.commit("base version")
        tree_b = work_a.controldir.sprout("wtb").open_workingtree()
        preview = work_a.basis_tree().preview_transform()
        self.addCleanup(preview.finalize)
        trans_id = preview.trans_id_tree_path("file")
        preview.delete_contents(trans_id)
        preview.create_file([b"b\nc\nd\ne\n"], trans_id)
        self.build_tree_contents([("wtb/file", b"a\nc\nd\nf\n")])
        tree_a = preview.get_preview_tree()
        if not getattr(tree_a, "plan_file_merge", None):
            self.skipTest("tree does not support file merge planning")
        tree_a.set_parent_ids([base_id])
        self.addCleanup(tree_b.lock_read().unlock)
        self.assertEqual(
            [
                ("killed-a", b"a\n"),
                ("killed-b", b"b\n"),
                ("unchanged", b"c\n"),
                ("unchanged", b"d\n"),
                ("new-a", b"e\n"),
                ("new-b", b"f\n"),
            ],
            list(tree_a.plan_file_merge("file", tree_b)),
        )

    def test_walkdirs(self):
        preview = self.get_empty_preview()
        preview.new_directory("", ROOT_PARENT, b"tree-root")
        # FIXME: new_directory should mark root.
        preview.fixup_new_roots()
        preview_tree = preview.get_preview_tree()
        preview.new_file("a", preview.root, [b"contents"], b"a-id")
        expected = [("", [("a", "a", "file", None, "file")])]
        self.assertEqual(expected, list(preview_tree.walkdirs()))

    def test_extras(self):
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree(
            ["tree/removed-file", "tree/existing-file", "tree/not-removed-file"]
        )
        work_tree.add(["removed-file", "not-removed-file"])
        preview = work_tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.new_file("new-file", preview.root, [b"contents"])
        preview.new_file(
            "new-versioned-file", preview.root, [b"contents"], b"new-versioned-id"
        )
        tree = preview.get_preview_tree()
        self.assertEqual({"existing-file"}, set(work_tree.extras()))
        preview.unversion_file(preview.trans_id_tree_path("removed-file"))
        self.assertEqual(
            {"new-file", "removed-file", "existing-file"}, set(tree.extras())
        )

    def test_merge_into_preview(self):
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", b"b\n")])
        work_tree.add("file")
        work_tree.commit("first commit")
        child_tree = work_tree.controldir.sprout("child").open_workingtree()
        self.build_tree_contents([("child/file", b"b\nc\n")])
        child_tree.commit("child commit")
        child_tree.lock_write()
        self.addCleanup(child_tree.unlock)
        work_tree.lock_write()
        self.addCleanup(work_tree.unlock)
        preview = work_tree.preview_transform()
        self.addCleanup(preview.finalize)
        file_trans_id = preview.trans_id_tree_path("file")
        preview.delete_contents(file_trans_id)
        preview.create_file([b"a\nb\n"], file_trans_id)
        preview_tree = preview.get_preview_tree()
        merger = Merger.from_revision_ids(
            preview_tree,
            child_tree.branch.last_revision(),
            other_branch=child_tree.branch,
            tree_branch=work_tree.branch,
        )
        merger.merge_type = Merge3Merger
        tt = merger.make_merger().make_preview_transform()
        self.addCleanup(tt.finalize)
        final_tree = tt.get_preview_tree()
        self.assertEqual(b"a\nb\nc\n", final_tree.get_file_text("file"))

    def test_merge_preview_into_workingtree(self):
        tree = self.make_branch_and_tree("tree")
        if tree.supports_setting_file_ids():
            tree.set_root_id(b"TREE_ROOT")
        tt = tree.preview_transform()
        self.addCleanup(tt.finalize)
        tt.new_file("name", tt.root, [b"content"], b"file-id")
        tree2 = self.make_branch_and_tree("tree2")
        if tree.supports_setting_file_ids():
            tree2.set_root_id(b"TREE_ROOT")
        merger = Merger.from_uncommitted(
            tree2, tt.get_preview_tree(), tree.basis_tree()
        )
        merger.merge_type = Merge3Merger
        merger.do_merge()

    def test_merge_preview_into_workingtree_handles_conflicts(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/foo", b"bar")])
        tree.add("foo")
        tree.commit("foo")
        tt = tree.preview_transform()
        self.addCleanup(tt.finalize)
        trans_id = tt.trans_id_tree_path("foo")
        tt.delete_contents(trans_id)
        tt.create_file([b"baz"], trans_id)
        tree2 = tree.controldir.sprout("tree2").open_workingtree()
        self.build_tree_contents([("tree2/foo", b"qux")])
        merger = Merger.from_uncommitted(
            tree2, tt.get_preview_tree(), tree.basis_tree()
        )
        merger.merge_type = Merge3Merger
        merger.do_merge()

    def test_has_filename(self):
        wt = self.make_branch_and_tree("tree")
        self.build_tree(["tree/unmodified", "tree/removed", "tree/modified"])
        tt = wt.preview_transform()
        removed_id = tt.trans_id_tree_path("removed")
        tt.delete_contents(removed_id)
        tt.new_file("new", tt.root, [b"contents"])
        modified_id = tt.trans_id_tree_path("modified")
        tt.delete_contents(modified_id)
        tt.create_file([b"modified-contents"], modified_id)
        self.addCleanup(tt.finalize)
        tree = tt.get_preview_tree()
        self.assertTrue(tree.has_filename("unmodified"))
        self.assertFalse(tree.has_filename("not-present"))
        self.assertFalse(tree.has_filename("removed"))
        self.assertTrue(tree.has_filename("new"))
        self.assertTrue(tree.has_filename("modified"))

    def test_is_executable2(self):
        tree = self.make_branch_and_tree("tree")
        preview = tree.preview_transform()
        self.addCleanup(preview.finalize)
        preview.new_file("foo", preview.root, [b"bar"], b"baz-id")
        preview_tree = preview.get_preview_tree()
        self.assertEqual(False, preview_tree.is_executable("tree/foo"))

    def test_commit_preview_tree(self):
        tree = self.make_branch_and_tree("tree")
        rev_id = tree.commit("rev1")
        tree.branch.lock_write()
        self.addCleanup(tree.branch.unlock)
        tt = tree.preview_transform()
        tt.new_file("file", tt.root, [b"contents"], b"file_id")
        self.addCleanup(tt.finalize)
        preview = tt.get_preview_tree()
        preview.set_parent_ids([rev_id])
        builder = tree.branch.get_commit_builder([rev_id])
        list(builder.record_iter_changes(preview, rev_id, tt.iter_changes()))
        builder.finish_inventory()
        rev2_id = builder.commit("rev2")
        rev2_tree = tree.branch.repository.revision_tree(rev2_id)
        self.assertEqual(b"contents", rev2_tree.get_file_text("file"))

    def test_ascii_limbo_paths(self):
        self.requireFeature(UnicodeFilenameFeature)
        branch = self.make_branch("any")
        tree = branch.repository.revision_tree(_mod_revision.NULL_REVISION)
        tt = tree.preview_transform()
        self.addCleanup(tt.finalize)
        foo_id = tt.new_directory("", ROOT_PARENT)
        bar_id = tt.new_file("\u1234bar", foo_id, [b"contents"])
        limbo_path = tt._limbo_name(bar_id)
        self.assertEqual(limbo_path, limbo_path)



# Copyright (C) 2006, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that lazy regexes are not compiled right away."""

import pickle
import re

from .. import lazy_regex, tests


class TestErrors(tests.TestCase):
    def test_invalid_pattern(self):
        error = lazy_regex.InvalidPattern("Bad pattern msg.")
        self.assertEqualDiff("Invalid pattern(s) found. Bad pattern msg.", str(error))


class InstrumentedLazyRegex(lazy_regex.LazyRegex):
    """Keep track of actions on the lazy regex."""

    _actions: list[tuple[str, ...]] = []

    @classmethod
    def use_actions(cls, actions):
        cls._actions = actions

    def __getattr__(self, attr):
        self._actions.append(("__getattr__", attr))
        return super().__getattr__(attr)

    def _real_re_compile(self, *args, **kwargs):
        self._actions.append(("_real_re_compile", args, kwargs))
        return super()._real_re_compile(*args, **kwargs)


class TestLazyRegex(tests.TestCase):
    def test_lazy_compile(self):
        """Make sure that LazyRegex objects compile at the right time."""
        actions = []
        InstrumentedLazyRegex.use_actions(actions)

        pattern = InstrumentedLazyRegex(args=("foo",), kwargs={})
        actions.append(("created regex", "foo"))
        # This match call should compile the regex and go through __getattr__
        pattern.match("foo")
        # But a further call should not go through __getattr__ because it has
        # been bound locally.
        pattern.match("foo")

        self.assertEqual(
            [
                ("created regex", "foo"),
                ("__getattr__", "match"),
                ("_real_re_compile", ("foo",), {}),
            ],
            actions,
        )

    def test_bad_pattern(self):
        """Ensure lazy regex handles bad patterns cleanly."""
        p = lazy_regex.lazy_compile("RE:[")
        # As p.match is lazy, we make it into a lambda so its handled
        # by assertRaises correctly.
        e = self.assertRaises(lazy_regex.InvalidPattern, lambda: p.match("foo"))
        # Expect either old or new form of error message
        self.assertContainsRe(
            e.msg,
            '^"RE:\\[" '
            "(unexpected end of regular expression"
            "|unterminated character set at position 3)$",
        )


class TestLazyCompile(tests.TestCase):
    def test_simple_acts_like_regex(self):
        """Test that the returned object has basic regex like functionality."""
        pattern = lazy_regex.lazy_compile("foo")
        self.assertIsInstance(pattern, lazy_regex.LazyRegex)
        self.assertTrue(pattern.match("foo"))
        self.assertIs(None, pattern.match("bar"))

    def test_extra_args(self):
        """Test that extra arguments are also properly passed."""
        pattern = lazy_regex.lazy_compile("foo", re.I)
        self.assertIsInstance(pattern, lazy_regex.LazyRegex)
        self.assertTrue(pattern.match("foo"))
        self.assertTrue(pattern.match("Foo"))

    def test_findall(self):
        pattern = lazy_regex.lazy_compile("fo*")
        self.assertEqual(["f", "fo", "foo", "fooo"], pattern.findall("f fo foo fooo"))

    def test_finditer(self):
        pattern = lazy_regex.lazy_compile("fo*")
        matches = [
            (m.start(), m.end(), m.group()) for m in pattern.finditer("foo bar fop")
        ]
        self.assertEqual([(0, 3, "foo"), (8, 10, "fo")], matches)

    def test_match(self):
        pattern = lazy_regex.lazy_compile("fo*")
        self.assertIs(None, pattern.match("baz foo"))
        self.assertEqual("fooo", pattern.match("fooo").group())

    def test_search(self):
        pattern = lazy_regex.lazy_compile("fo*")
        self.assertEqual("foo", pattern.search("baz foo").group())
        self.assertEqual("fooo", pattern.search("fooo").group())

    def test_split(self):
        pattern = lazy_regex.lazy_compile("[,;]+")
        self.assertEqual(["x", "y", "z"], pattern.split("x,y;z"))

    def test_pickle(self):
        # When pickling, just compile the regex.
        # Sphinx, which we use for documentation, pickles
        # some compiled regexes.
        lazy_pattern = lazy_regex.lazy_compile("[,;]+")
        pickled = pickle.dumps(lazy_pattern)
        unpickled_lazy_pattern = pickle.loads(pickled)
        self.assertEqual(["x", "y", "z"], unpickled_lazy_pattern.split("x,y;z"))



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz diff."""

import os
import re

from breezy import tests, workingtree
from breezy.diff import DiffTree
from breezy.diff import format_registry as diff_format_registry
from breezy.tests import features


def subst_dates(string):
    """Replace date strings with constant values."""
    return re.sub(
        r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2} [-\+]\d{4}",
        "YYYY-MM-DD HH:MM:SS +ZZZZ",
        string,
    )


class DiffBase(tests.TestCaseWithTransport):
    """Base class with common setup method."""

    def make_example_branch(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("hello", b"foo\n"), ("goodbye", b"baz\n")])
        tree.add(["hello"])
        tree.commit("setup")
        tree.add(["goodbye"])
        tree.commit("setup")
        return tree


class TestDiff(DiffBase):
    def test_diff(self):
        tree = self.make_example_branch()
        self.build_tree_contents([("hello", b"hello world!")])
        tree.commit(message="fixing hello")
        output = self.run_bzr("diff -r 2..3", retcode=1)[0]
        self.assertTrue("\n+hello world!" in output)
        output = self.run_bzr("diff -c 3", retcode=1)[0]
        self.assertTrue("\n+hello world!" in output)
        output = self.run_bzr("diff -r last:3..last:1", retcode=1)[0]
        self.assertTrue("\n+baz" in output)
        output = self.run_bzr("diff -c last:2", retcode=1)[0]
        self.assertTrue("\n+baz" in output)
        self.build_tree(["moo"])
        tree.add("moo")
        os.unlink("moo")
        self.run_bzr("diff")

    def test_diff_prefix(self):
        """Diff --prefix appends to filenames in output."""
        self.make_example_branch()
        self.build_tree_contents([("hello", b"hello world!\n")])
        out, err = self.run_bzr("diff --prefix old/:new/", retcode=1)
        self.assertEqual(err, "")
        self.assertEqualDiff(
            subst_dates(out),
            """\
=== modified file 'hello'
--- old/hello\tYYYY-MM-DD HH:MM:SS +ZZZZ
+++ new/hello\tYYYY-MM-DD HH:MM:SS +ZZZZ
@@ -1,1 +1,1 @@
-foo
+hello world!

""",
        )

    def test_diff_illegal_prefix_value(self):
        # There was an error in error reporting for this option
        out, err = self.run_bzr("diff --prefix old/", retcode=3)
        self.assertContainsRe(err, "--prefix expects two values separated by a colon")

    def test_diff_p1(self):
        """Diff -p1 produces lkml-style diffs."""
        self.make_example_branch()
        self.build_tree_contents([("hello", b"hello world!\n")])
        out, err = self.run_bzr("diff -p1", retcode=1)
        self.assertEqual(err, "")
        self.assertEqualDiff(
            subst_dates(out),
            """\
=== modified file 'hello'
--- old/hello\tYYYY-MM-DD HH:MM:SS +ZZZZ
+++ new/hello\tYYYY-MM-DD HH:MM:SS +ZZZZ
@@ -1,1 +1,1 @@
-foo
+hello world!

""",
        )

    def test_diff_p0(self):
        """Diff -p0 produces diffs with no prefix."""
        self.make_example_branch()
        self.build_tree_contents([("hello", b"hello world!\n")])
        out, err = self.run_bzr("diff -p0", retcode=1)
        self.assertEqual(err, "")
        self.assertEqualDiff(
            subst_dates(out),
            """\
=== modified file 'hello'
--- hello\tYYYY-MM-DD HH:MM:SS +ZZZZ
+++ hello\tYYYY-MM-DD HH:MM:SS +ZZZZ
@@ -1,1 +1,1 @@
-foo
+hello world!

""",
        )

    def test_diff_nonexistent(self):
        # Get an error from a file that does not exist at all
        # (Malone #3619)
        self.make_example_branch()
        out, err = self.run_bzr(
            "diff does-not-exist",
            retcode=3,
            error_regexes=("not versioned.*does-not-exist",),
        )

    def test_diff_illegal_revision_specifiers(self):
        out, err = self.run_bzr(
            "diff -r 1..23..123",
            retcode=3,
            error_regexes=("one or two revision specifiers",),
        )

    def test_diff_using_and_format(self):
        out, err = self.run_bzr(
            "diff --format=default --using=mydi",
            retcode=3,
            error_regexes=("are mutually exclusive",),
        )

    def test_diff_nonexistent_revision(self):
        out, err = self.run_bzr(
            "diff -r 123",
            retcode=3,
            error_regexes=("Requested revision: '123' does not exist in branch:",),
        )

    def test_diff_nonexistent_dotted_revision(self):
        out, err = self.run_bzr("diff -r 1.1", retcode=3)
        self.assertContainsRe(
            err, "Requested revision: '1.1' does not exist in branch:"
        )

    def test_diff_nonexistent_dotted_revision_change(self):
        out, err = self.run_bzr("diff -c 1.1", retcode=3)
        self.assertContainsRe(
            err, "Requested revision: '1.1' does not exist in branch:"
        )

    def test_diff_unversioned(self):
        # Get an error when diffing a non-versioned file.
        # (Malone #3619)
        self.make_example_branch()
        self.build_tree(["unversioned-file"])
        out, err = self.run_bzr("diff unversioned-file", retcode=3)
        self.assertContainsRe(err, "not versioned.*unversioned-file")

    # TODO: What should diff say for a file deleted in working tree?

    def example_branches(self):
        branch1_tree = self.make_branch_and_tree("branch1")
        self.build_tree(["branch1/file"], line_endings="binary")
        self.build_tree(["branch1/file2"], line_endings="binary")
        branch1_tree.add("file")
        branch1_tree.add("file2")
        branch1_tree.commit(message="add file and file2")
        branch2_tree = branch1_tree.controldir.sprout("branch2").open_workingtree()
        self.build_tree_contents([("branch2/file", b"new content\n")])
        branch2_tree.commit(message="update file")
        return branch1_tree, branch2_tree

    def check_b2_vs_b1(self, cmd):
        # Compare branch2 vs branch1 using cmd and check the result
        out, err = self.run_bzr(cmd, retcode=1)
        self.assertEqual("", err)
        self.assertEqual(
            "=== modified file 'file'\n"
            "--- old/file\tYYYY-MM-DD HH:MM:SS +ZZZZ\n"
            "+++ new/file\tYYYY-MM-DD HH:MM:SS +ZZZZ\n"
            "@@ -1,1 +1,1 @@\n"
            "-new content\n"
            "+contents of branch1/file\n"
            "\n",
            subst_dates(out),
        )

    def check_b1_vs_b2(self, cmd):
        # Compare branch1 vs branch2 using cmd and check the result
        out, err = self.run_bzr(cmd, retcode=1)
        self.assertEqual("", err)
        self.assertEqualDiff(
            "=== modified file 'file'\n"
            "--- old/file\tYYYY-MM-DD HH:MM:SS +ZZZZ\n"
            "+++ new/file\tYYYY-MM-DD HH:MM:SS +ZZZZ\n"
            "@@ -1,1 +1,1 @@\n"
            "-contents of branch1/file\n"
            "+new content\n"
            "\n",
            subst_dates(out),
        )

    def check_no_diffs(self, cmd):
        # Check that running cmd returns an empty diff
        out, err = self.run_bzr(cmd, retcode=0)
        self.assertEqual("", err)
        self.assertEqual("", out)

    def test_diff_branches(self):
        self.example_branches()
        # should open branch1 and diff against branch2,
        self.check_b2_vs_b1("diff -r branch:branch2 branch1")
        # Compare two working trees using various syntax forms
        self.check_b2_vs_b1("diff --old branch2 --new branch1")
        self.check_b2_vs_b1("diff --old branch2 branch1")
        self.check_b2_vs_b1("diff branch2 --new branch1")
        # Test with a selected file that was changed
        self.check_b2_vs_b1("diff --old branch2 --new branch1 file")
        self.check_b2_vs_b1("diff --old branch2 branch1/file")
        self.check_b2_vs_b1("diff branch2/file --new branch1")
        # Test with a selected file that was not changed
        self.check_no_diffs("diff --old branch2 --new branch1 file2")
        self.check_no_diffs("diff --old branch2 branch1/file2")
        self.check_no_diffs("diff branch2/file2 --new branch1")

    def test_diff_branches_no_working_trees(self):
        branch1_tree, branch2_tree = self.example_branches()
        # Compare a working tree to a branch without a WT
        dir1 = branch1_tree.controldir
        dir1.destroy_workingtree()
        self.assertFalse(dir1.has_workingtree())
        self.check_b2_vs_b1("diff --old branch2 --new branch1")
        self.check_b2_vs_b1("diff --old branch2 branch1")
        self.check_b2_vs_b1("diff branch2 --new branch1")
        # Compare a branch without a WT to one with a WT
        self.check_b1_vs_b2("diff --old branch1 --new branch2")
        self.check_b1_vs_b2("diff --old branch1 branch2")
        self.check_b1_vs_b2("diff branch1 --new branch2")
        # Compare a branch with a WT against another without a WT
        dir2 = branch2_tree.controldir
        dir2.destroy_workingtree()
        self.assertFalse(dir2.has_workingtree())
        self.check_b1_vs_b2("diff --old branch1 --new branch2")
        self.check_b1_vs_b2("diff --old branch1 branch2")
        self.check_b1_vs_b2("diff branch1 --new branch2")

    def test_diff_revno_branches(self):
        self.example_branches()
        branch2_tree = workingtree.WorkingTree.open_containing("branch2")[0]
        self.build_tree_contents([("branch2/file", b"even newer content")])
        branch2_tree.commit(message="update file once more")

        out, err = self.run_bzr(
            "diff -r revno:1:branch2..revno:1:branch1",
        )
        self.assertEqual("", err)
        self.assertEqual("", out)
        out, err = self.run_bzr("diff -r revno:2:branch2..revno:1:branch1", retcode=1)
        self.assertEqual("", err)
        self.assertEqualDiff(
            "=== modified file 'file'\n"
            "--- old/file\tYYYY-MM-DD HH:MM:SS +ZZZZ\n"
            "+++ new/file\tYYYY-MM-DD HH:MM:SS +ZZZZ\n"
            "@@ -1,1 +1,1 @@\n"
            "-new content\n"
            "+contents of branch1/file\n"
            "\n",
            subst_dates(out),
        )

    def test_diff_color_always(self):
        from ... import colordiff
        from ...terminal import colorstring

        self.overrideAttr(colordiff, "GLOBAL_COLORDIFFRC", None)
        self.example_branches()
        branch2_tree = workingtree.WorkingTree.open_containing("branch2")[0]
        self.build_tree_contents([("branch2/file", b"even newer content")])
        branch2_tree.commit(message="update file once more")

        out, err = self.run_bzr(
            "diff --color=always -r revno:2:branch2..revno:1:branch1", retcode=1
        )
        self.assertEqual("", err)
        self.assertEqualDiff(
            (
                colorstring(b"=== modified file 'file'\n", "darkyellow")
                + colorstring(b"--- old/file\tYYYY-MM-DD HH:MM:SS +ZZZZ\n", "darkred")
                + colorstring(b"+++ new/file\tYYYY-MM-DD HH:MM:SS +ZZZZ\n", "darkblue")
                + colorstring(b"@@ -1 +1 @@\n", "darkgreen")
                + colorstring(b"-new content\n", "darkred")
                + colorstring(b"+contents of branch1/file\n", "darkblue")
                + colorstring(b"\n", "darkwhite")
            ).decode(),
            subst_dates(out),
        )

    def example_branch2(self):
        branch1_tree = self.make_branch_and_tree("branch1")
        self.build_tree_contents([("branch1/file1", b"original line\n")])
        branch1_tree.add("file1")
        branch1_tree.commit(message="first commit")
        self.build_tree_contents([("branch1/file1", b"repo line\n")])
        branch1_tree.commit(message="second commit")
        return branch1_tree

    def test_diff_to_working_tree(self):
        self.example_branch2()
        self.build_tree_contents([("branch1/file1", b"new line")])
        output = self.run_bzr("diff -r 1.. branch1", retcode=1)
        self.assertContainsRe(output[0], "\n\\-original line\n\\+new line\n")

    def test_diff_to_working_tree_in_subdir(self):
        self.example_branch2()
        self.build_tree_contents([("branch1/file1", b"new line")])
        os.mkdir("branch1/dir1")
        output = self.run_bzr("diff -r 1..", retcode=1, working_dir="branch1/dir1")
        self.assertContainsRe(output[0], "\n\\-original line\n\\+new line\n")

    def test_diff_across_rename(self):
        """The working tree path should always be considered for diffing."""
        tree = self.make_example_branch()
        self.run_bzr("diff -r 0..1 hello", retcode=1)
        tree.rename_one("hello", "hello1")
        self.run_bzr("diff hello1", retcode=1)
        self.run_bzr("diff -r 0..1 hello1", retcode=1)

    def test_diff_to_branch_no_working_tree(self):
        branch1_tree = self.example_branch2()
        dir1 = branch1_tree.controldir
        dir1.destroy_workingtree()
        self.assertFalse(dir1.has_workingtree())
        output = self.run_bzr("diff -r 1.. branch1", retcode=1)
        self.assertContainsRe(output[0], "\n\\-original line\n\\+repo line\n")

    def test_custom_format(self):
        class BooDiffTree(DiffTree):
            def show_diff(self, specific_files, extra_trees=None):
                self.to_file.write("BOO!\n")
                return super().show_diff(specific_files, extra_trees)

        diff_format_registry.register("boo", BooDiffTree, "Scary diff format")
        self.addCleanup(diff_format_registry.remove, "boo")
        self.make_example_branch()
        self.build_tree_contents([("hello", b"hello world!\n")])
        output = self.run_bzr("diff --format=boo", retcode=1)
        self.assertTrue("BOO!" in output[0])
        output = self.run_bzr("diff -Fboo", retcode=1)
        self.assertTrue("BOO!" in output[0])

    def test_binary_diff_remove(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("a", b"\x00" * 20)])
        tree.add(["a"])
        tree.commit("add binary file")
        os.unlink("a")
        output = self.run_bzr("diff", retcode=1)
        self.assertEqual(
            "=== removed file 'a'\nBinary files old/a and new/a differ\n", output[0]
        )

    def test_moved_away(self):
        # pad.lv/1880354
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("a", "asdf\n")])
        tree.add(["a"])
        tree.commit("add a")
        tree.rename_one("a", "b")
        self.build_tree_contents([("a", "qwer\n")])
        tree.add("a")
        output, error = self.run_bzr("diff -p0", retcode=1)
        self.assertEqualDiff(
            """\
=== added file 'a'
--- a\tYYYY-MM-DD HH:MM:SS +ZZZZ
+++ a\tYYYY-MM-DD HH:MM:SS +ZZZZ
@@ -0,0 +1,1 @@
+qwer

=== renamed file 'a' => 'b'
""",
            subst_dates(output),
        )


class TestCheckoutDiff(TestDiff):
    def make_example_branch(self):
        tree = super().make_example_branch()
        tree = tree.branch.create_checkout("checkout")
        os.chdir("checkout")
        return tree

    def example_branch2(self):
        tree = super().example_branch2()
        os.mkdir("checkouts")
        tree = tree.branch.create_checkout("checkouts/branch1")
        os.chdir("checkouts")
        return tree

    def example_branches(self):
        branch1_tree, branch2_tree = super().example_branches()
        os.mkdir("checkouts")
        branch1_tree = branch1_tree.branch.create_checkout("checkouts/branch1")
        branch2_tree = branch2_tree.branch.create_checkout("checkouts/branch2")
        os.chdir("checkouts")
        return branch1_tree, branch2_tree


class TestDiffLabels(DiffBase):
    def test_diff_label_removed(self):
        tree = super().make_example_branch()
        tree.remove("hello", keep_files=False)
        diff = self.run_bzr("diff", retcode=1)
        self.assertTrue("=== removed file 'hello'" in diff[0])

    def test_diff_label_added(self):
        tree = super().make_example_branch()
        self.build_tree_contents([("barbar", b"barbar")])
        tree.add("barbar")
        diff = self.run_bzr("diff", retcode=1)
        self.assertTrue("=== added file 'barbar'" in diff[0])

    def test_diff_label_modified(self):
        super().make_example_branch()
        self.build_tree_contents([("hello", b"barbar")])
        diff = self.run_bzr("diff", retcode=1)
        self.assertTrue("=== modified file 'hello'" in diff[0])

    def test_diff_label_renamed(self):
        tree = super().make_example_branch()
        tree.rename_one("hello", "gruezi")
        diff = self.run_bzr("diff", retcode=1)
        self.assertTrue("=== renamed file 'hello' => 'gruezi'" in diff[0])


class TestExternalDiff(DiffBase):
    def test_external_diff(self):
        """Test that we can spawn an external diff process."""
        self.disable_missing_extensions_warning()
        # We have to use run_brz_subprocess, because we need to
        # test writing directly to stdout, (there was a bug in
        # subprocess.py that we had to workaround).
        # However, if 'diff' may not be available
        self.make_example_branch()
        out, err = self.run_brz_subprocess(
            "diff -Oprogress_bar=none -r 1 --diff-options -ub",
            universal_newlines=True,
            retcode=None,
        )
        if b"Diff is not installed on this machine" in err:
            raise tests.TestSkipped("No external 'diff' is available")
        self.assertEqual(b"", err)
        # We have to skip the stuff in the middle, because it depends
        # on time.time()
        self.assertStartsWith(
            out,
            b"=== added file 'goodbye'\n"
            b"--- old/goodbye\t1970-01-01 00:00:00 +0000\n"
            b"+++ new/goodbye\t",
        )
        self.assertEndsWith(out, b"\n@@ -0,0 +1 @@\n+baz\n\n")

    def test_external_diff_options_and_using(self):
        """Test that the options are passed correctly to an external diff process."""
        self.requireFeature(features.diff_feature)
        self.make_example_branch()
        self.build_tree_contents([("hello", b"Foo\n")])
        out, err = self.run_bzr(
            "diff --diff-options -i --diff-options -a --using diff", retcode=1
        )
        self.assertEqual("=== modified file 'hello'\n1c1\n< foo\n---\n> Foo\n", out)
        self.assertEqual("", err)


class TestDiffOutput(DiffBase):
    def test_diff_output(self):
        # check that output doesn't mangle line-endings
        self.make_example_branch()
        self.build_tree_contents([("hello", b"hello world!\n")])
        output = self.run_brz_subprocess("diff", retcode=1)[0]
        self.assertTrue(b"\n+hello world!\n" in output)



#!/usr/bin/env python3

import errno
import sys
from io import StringIO

try:
    from docutils.core import publish_file
    from docutils.parsers import rst
except ModuleNotFoundError:
    print("Missing dependency.  Please install docutils.")
    sys.exit(1)
try:
    from elementtree import HTMLTreeBuilder
    from elementtree.ElementTree import XML
except ModuleNotFoundError:
    print("Missing dependency.  Please install ElementTree.")
    sys.exit(1)
try:
    import kid
except ModuleNotFoundError:
    print("Missing dependency.  Please install Kid.")
    sys.exit(1)


def kidified_rest(rest_file, template_name):
    xhtml_file = StringIO()
    # prevent docutils from autoclosing the StringIO
    xhtml_file.close = lambda: None
    publish_file(
        rest_file,
        writer_name="html",
        destination=xhtml_file,
        settings_overrides={"doctitle_xform": 0},
    )
    xhtml_file.seek(0)
    xml = HTMLTreeBuilder.parse(xhtml_file)
    head = xml.find("head")
    body = xml.find("body")
    assert head is not None
    assert body is not None
    template = kid.Template(file=template_name, head=head, body=body)
    return template.serialize(output="html")


def safe_open(filename, mode):
    try:
        return open(filename, mode + "b")
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        sys.stderr.write("file not found: {}\n".format(sys.argv[2]))
        sys.exit(3)


def main(template, source=None, target=None):
    if source is not None:
        rest_file = safe_open(source, "r")
    else:
        rest_file = sys.stdin
    if target is not None:
        out_file = safe_open(target, "w")
    else:
        out_file = sys.stdout
    out_file.write(kidified_rest(rest_file, template))


assert len(sys.argv) > 1

# Strip options so only the arguments are passed
args = [x for x in sys.argv[1:] if not x.startswith("-")]
main(*args)



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import sys
import tempfile

from .. import mergetools, tests


class TestFilenameSubstitution(tests.TestCaseInTempDir):
    def test_simple_filename(self):
        cmd_list = ["kdiff3", "{base}", "{this}", "{other}", "-o", "{result}"]
        args, tmpfile = mergetools._subst_filename(cmd_list, "test.txt")
        self.assertEqual(
            [
                "kdiff3",
                "test.txt.BASE",
                "test.txt.THIS",
                "test.txt.OTHER",
                "-o",
                "test.txt",
            ],
            args,
        )

    def test_spaces(self):
        cmd_list = ["kdiff3", "{base}", "{this}", "{other}", "-o", "{result}"]
        args, tmpfile = mergetools._subst_filename(cmd_list, "file with space.txt")
        self.assertEqual(
            [
                "kdiff3",
                "file with space.txt.BASE",
                "file with space.txt.THIS",
                "file with space.txt.OTHER",
                "-o",
                "file with space.txt",
            ],
            args,
        )

    def test_spaces_and_quotes(self):
        cmd_list = ["kdiff3", "{base}", "{this}", "{other}", "-o", "{result}"]
        args, tmpfile = mergetools._subst_filename(
            cmd_list, 'file with "space and quotes".txt'
        )
        self.assertEqual(
            [
                "kdiff3",
                'file with "space and quotes".txt.BASE',
                'file with "space and quotes".txt.THIS',
                'file with "space and quotes".txt.OTHER',
                "-o",
                'file with "space and quotes".txt',
            ],
            args,
        )

    def test_tempfile(self):
        self.build_tree(
            ("test.txt", "test.txt.BASE", "test.txt.THIS", "test.txt.OTHER")
        )
        cmd_list = ["some_tool", "{this_temp}"]
        args, tmpfile = mergetools._subst_filename(cmd_list, "test.txt")
        self.assertPathExists(tmpfile)
        os.remove(tmpfile)


class TestCheckAvailability(tests.TestCaseInTempDir):
    def test_full_path(self):
        self.assertTrue(mergetools.check_availability(sys.executable))

    def test_nonexistent(self):
        self.assertFalse(mergetools.check_availability("DOES NOT EXIST"))

    def test_non_executable(self):
        f, name = tempfile.mkstemp()
        try:
            self.log("temp filename: %s", name)
            self.assertFalse(mergetools.check_availability(name))
        finally:
            os.close(f)
            os.unlink(name)


class TestInvoke(tests.TestCaseInTempDir):
    def setUp(self):
        super(tests.TestCaseInTempDir, self).setUp()
        self._exe = None
        self._args = None
        self.build_tree_contents(
            (
                ("test.txt", b"stuff"),
                ("test.txt.BASE", b"base stuff"),
                ("test.txt.THIS", b"this stuff"),
                ("test.txt.OTHER", b"other stuff"),
            )
        )

    def test_invoke_expands_exe_path(self):
        self.overrideEnv("PATH", os.path.dirname(sys.executable))

        def dummy_invoker(exe, args, cleanup):
            self._exe = exe
            self._args = args
            cleanup(0)
            return 0

        command = "{} {{result}}".format(os.path.basename(sys.executable))
        retcode = mergetools.invoke(command, "test.txt", dummy_invoker)
        self.assertEqual(0, retcode)
        self.assertEqual(sys.executable, self._exe)
        self.assertEqual(["test.txt"], self._args)

    def test_success(self):
        def dummy_invoker(exe, args, cleanup):
            self._exe = exe
            self._args = args
            cleanup(0)
            return 0

        retcode = mergetools.invoke("tool {result}", "test.txt", dummy_invoker)
        self.assertEqual(0, retcode)
        self.assertEqual("tool", self._exe)
        self.assertEqual(["test.txt"], self._args)

    def test_failure(self):
        def dummy_invoker(exe, args, cleanup):
            self._exe = exe
            self._args = args
            cleanup(1)
            return 1

        retcode = mergetools.invoke("tool {result}", "test.txt", dummy_invoker)
        self.assertEqual(1, retcode)
        self.assertEqual("tool", self._exe)
        self.assertEqual(["test.txt"], self._args)

    def test_success_tempfile(self):
        def dummy_invoker(exe, args, cleanup):
            self._exe = exe
            self._args = args
            self.assertPathExists(args[0])
            with open(args[0], "w") as f:
                f.write("temp stuff")
            cleanup(0)
            return 0

        retcode = mergetools.invoke("tool {this_temp}", "test.txt", dummy_invoker)
        self.assertEqual(0, retcode)
        self.assertEqual("tool", self._exe)
        self.assertPathDoesNotExist(self._args[0])
        self.assertFileEqual(b"temp stuff", "test.txt")

    def test_failure_tempfile(self):
        def dummy_invoker(exe, args, cleanup):
            self._exe = exe
            self._args = args
            self.assertPathExists(args[0])
            self.log(repr(args))
            with open(args[0], "w") as f:
                self.log(repr(f))
                f.write("temp stuff")
            cleanup(1)
            return 1

        retcode = mergetools.invoke("tool {this_temp}", "test.txt", dummy_invoker)
        self.assertEqual(1, retcode)
        self.assertEqual("tool", self._exe)
        self.assertFileEqual(b"stuff", "test.txt")



# Copyright (C) 2010-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test the git revision specifiers."""

from ...tests import TestCase
from ..revspec import valid_git_sha1


class Sha1ValidTests(TestCase):
    def test_invalid(self):
        self.assertFalse(valid_git_sha1(b"git-v1:abcde"))

    def test_valid(self):
        self.assertTrue(valid_git_sha1(b"aabbccddee"))
        self.assertTrue(valid_git_sha1(b"aabbccd"))



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""OS Lock implementation tests for bzr.

These test the conformance of all the lock variations to the expected API.
"""

from breezy import lock, tests


class TestCaseWithLock(tests.TestCaseWithTransport):
    pass


def make_scenarios(lock_classes):
    result = []
    for name, write_lock, read_lock in lock_classes:
        result.append((name, {"write_lock": write_lock, "read_lock": read_lock}))
    return result


def load_tests(loader, standard_tests, pattern):
    submod_tests = loader.loadTestsFromModuleNames(
        [
            "breezy.tests.per_lock.test_lock",
            "breezy.tests.per_lock.test_temporary_write_lock",
        ]
    )
    scenarios = make_scenarios(lock._lock_classes)
    # add the tests for the sub modules
    return tests.multiply_tests(submod_tests, scenarios, standard_tests)
